# -*- coding: utf-8 -*-
import json
import subprocess
import iscc_core.code_content_text
from iscc_core.code_content_text import Text
from loguru import logger as log
from os.path import basename, splitext
from typing import Any, Generator, Optional, Union
from urllib.parse import urlparse
import langdetect
import xxhash
from functools import lru_cache
import langcodes
import iscc.bin
from iscc.schema import FeatureType
from iscc_core.cdc import data_chunks
from iscc.options import SdkOptions, sdk_opts
from iscc import uread
from iscc.utils import sliding_window
from iscc_core.codec import encode_base64
from iscc.schema import Readable
from iscc_core.minhash import minhash_64, minhash_256


# Set for deterministic language detection
langdetect.DetectorFactory.seed = 0


def extract_text(data):
    # type: (Readable) -> str
    """Extract plaintext from a text document."""

    ufile = uread.open_data(data)
    cmd = [
        iscc.bin.java_bin(),
        "-jar",
        iscc.bin.tika_bin(),
        "--text",
        "--encoding=UTF-8",
    ]
    if hasattr(ufile, "name"):
        cmd.append(ufile.name)
        data = None
    else:
        data = ufile.read()
        if not data:
            log.warning(f"No data to extract text from {type(data)}")
            return ""

    try:
        result = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError:
        iscc.bin.tika_install()
        result = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)

    return result.stdout.decode(encoding="UTF-8")


def extract_text_metadata(data, text=None, **options):
    # type: (Readable, Optional[str], **Any) -> dict
    """Extract metadata from text document (title, language, characters).

    :param Readable data: Readable with textual content
    :param str text: Extracted text
    :key bool text_guess_title: Guess title from content if not found in metadata.
    """
    opts = SdkOptions(**options) if options else sdk_opts
    ufile = uread.open_data(data)
    cmd = [
        iscc.bin.java_bin(),
        "-jar",
        iscc.bin.tika_bin(),
        "--metadata",
        "-j",
        "--encoding=UTF-8",
    ]

    if hasattr(ufile, "name"):
        cmd.append(ufile.name)
        data = None
    else:
        data = ufile.read()
        if not data:
            log.warning(f"No data to extract text metadata from {type(data)}")
            return {"characters": 0}

    try:
        result = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError:
        iscc.bin.tika_install()
        result = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)

    try:
        metadata = json.loads(result.stdout)
    except json.JSONDecodeError:
        log.error("tike metadata json decode failed")
        metadata = {}

    title = metadata.get("dc:title", "")
    norm_title = normalize_text(title)

    # Falback to get title from content
    norm_text = normalize_text(text) if text else ""

    if not norm_title and opts.text_guess_title and text:
        title = text.strip().splitlines()[0]
        norm_title = normalize_text(title)

    norm_title = iscc_core.code_meta.trim_text(norm_title, opts.meta_trim_title)

    meta = {}
    if norm_title:
        meta["title"] = norm_title

    # Number of characters
    meta["characters"] = len(norm_text)

    try:
        lang = langdetect.detect(text)
        log.info(f"Detected langauge: {lang}")
        meta["language"] = langcodes.standardize_tag(lang)
    except Exception as e:
        log.warning(f"Language detection failed: {e}")

    return meta


def extract_text_features(text, **options):
    # type: (str, **Any) -> dict
    """
    Create granular fingerprint for text (minhashes over ngrams from cdc-chunks).
    Text should be normalized before extracting text features.

    :param str text: Normalized plaintext.
    :key text_avg_chunk_size: Avg. number of chars per text chunk to be hashed.
    :key text_ngram_size: Sliding window size in number of characters.
    :returns dict: Dictionary with 'sizes' and 'features'.
    """
    opts = SdkOptions(**options)
    text = text.lower()
    chunks = chunk_text(text, text_avg_chunk_size=opts.text_avg_chunk_size)
    sizes = []
    feats = []
    for chunk in chunks:
        ngrams = (
            "".join(chars) for chars in sliding_window(chunk, opts.text_ngram_size)
        )
        features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
        minimum_hash_digest = minhash_64(features)
        sizes.append(len(chunk))
        feats.append(encode_base64(minimum_hash_digest))
    return dict(kind=FeatureType.text.value, version=0, features=feats, sizes=sizes)


@lru_cache(maxsize=4, typed=True)
def normalize_text(text):
    # type: (Text) -> str
    """
    Apply Unicode normalization and character filtering.

    Wraps iscc_core.normalize_text to support caching

    - Decode bytes to Unicode (assuming UTF-8 encoded text).
    - Remove leading/trailing whitespace.
    - Decompose with NFD normalization.
    - Filter special characters and whitespace.
    - Remove duplicate whitespace.
    - Recombine with NFKC normalization.

    :param Text text: Plain text to be normalized.
    :return: Normalized plain text.
    :rtype: str
    """
    return iscc_core.code_content_text.normalize_text(text)


def hash_text(text, **options):
    # type: (str, **Any) -> bytes
    """
    Create a 256-bit similarity preserving hash for text input.
    Text should be normalized before hash creation.
    """
    opts = SdkOptions(**options)
    text = text.lower()
    ngrams = ("".join(chars) for chars in sliding_window(text, opts.text_ngram_size))
    features = [xxhash.xxh32_intdigest(s.encode("utf-8")) for s in ngrams]
    shash = minhash_256(features)
    return shash


def chunk_text(text, **options):
    # type: (str, **Any) -> Generator[str]
    """
    Generates variable sized text chunks (without leading BOM)

    :param text: normalized plaintext
    :key: text_avg_chunk_size: Targeted average size of text chunks in bytes.
    """
    opts = SdkOptions(**options)
    avg_size = opts.text_avg_chunk_size
    data = text.encode("utf-32-be")
    avg_size *= 4  # 4 bytes per character
    for chunk in data_chunks(data, utf32=True, avg_chunk_size=avg_size):
        yield chunk.decode("utf-32-be")


def trim_text(text, nbytes):
    # type: (str, int) -> str
    """Trim text such that its utf-8 encoded size does not exceed nbytes."""
    return text.encode("utf-8")[:nbytes].decode("utf-8", "ignore").strip()


def title_from_uri(uri):
    """Extract "filename" part of an uri without file extension to be uses as fallback
    title for an asset if no title information can be aquired.
    """
    result = urlparse(uri)
    base = basename(result.path)
    name = splitext(base)[0]
    name = name.replace("-", " ")
    name = name.replace("_", " ")
    return name



# -*- coding: utf-8 -*-
import io
from iscc.schema import Readable, Uri, File, Data
from iscc.utils import download_file
from tempfile import gettempdir


def open_data(data):
    # type: (Readable) -> Readable
    """Open filepath, uri, rawdata or file-like object."""
    if isinstance(data, Uri.__args__):
        if isinstance(data, str) and (
            data.startswith("http://") or data.startswith("https://")
        ):
            temp_folder = gettempdir()
            data = download_file(data, folder=temp_folder, sanitize=True)
        return open(str(data), "rb")
    elif isinstance(data, Data.__args__):
        return io.BytesIO(data)
    elif isinstance(data, File.__args__):
        data.seek(0)
        return data
    else:
        raise ValueError(f"unsupported data type {type(data)}")



# -*- coding: utf-8 -*-
import os
import platform
import shutil
import subprocess
import tarfile
import zipfile
from platform import system, architecture
import requests
from loguru import logger as log
import iscc
from iscc.utils import download_file
import stat
import jdk


FFPROBE_VERSION = "4.2.1"
FFPROBE_API_URL = "https://ffbinaries.com/api/v1/version/" + FFPROBE_VERSION
FFPROBE_CHECKSUMS = {
    "windows-64": "d13e513cbe25df56f2a003740dbfa3c9",
    "linux-64": "d16c0f22a4d86653f9a11ef9da74f644",
    "osx-64": "6cafd5ceede9d4e5bb2a241c9f22efd5",
}

FFMPEG_VERSION = "4.2.1"
FFMPEG_API_URL = "https://ffbinaries.com/api/v1/version/" + FFMPEG_VERSION
FFMPEG_CHECKSUMS = {
    "windows-64": "6d5dc6d2bb579ac3f78fda858f07bc5b",
    "linux-64": "01fb50166b77dd128dca786fe152ac2c",
    "osx-64": "c882209d862e33e9dc4f7abe1adb5e1b",
}

FPCALC_VERSION = "1.5.0"
FPCALC_URL_BASE = (
    f"https://github.com/acoustid/chromaprint/releases/download/v{FPCALC_VERSION}/"
)

FPCALC_OS_MAP = {
    "Linux": "chromaprint-fpcalc-{}-linux-x86_64.tar.gz".format(FPCALC_VERSION),
    "Darwin": "chromaprint-fpcalc-{}-macos-x86_64.tar.gz".format(FPCALC_VERSION),
    "Windows": "chromaprint-fpcalc-{}-windows-x86_64.zip".format(FPCALC_VERSION),
}

TIKA_VERSION = "2.2.0"
TIKA_URL = f"https://dlcdn.apache.org/tika/{TIKA_VERSION}/tika-app-{TIKA_VERSION}.jar"
TIKA_CHECKSUM = "62ba528230481c9917f2dcbf3524e63d"


def install():
    """Install binary tools for content extraction"""
    fpcalc_install()
    ffprobe_install()
    ffmpeg_install()
    java_install()
    tika_install()


def fpcalc_bin():
    """Returns local path to fpcalc executable."""
    if platform.system() == "Windows":
        return os.path.join(iscc.APP_DIR, "fpcalc-{}.exe".format(FPCALC_VERSION))
    return os.path.join(iscc.APP_DIR, "fpcalc-{}".format(FPCALC_VERSION))


def fpcalc_is_installed():
    """"Check if fpcalc is installed."""
    fp = fpcalc_bin()
    return os.path.isfile(fp) and os.access(fp, os.X_OK)


def fpcalc_download_url():
    """Return system and version dependant download url"""
    return os.path.join(FPCALC_URL_BASE, FPCALC_OS_MAP[platform.system()])


def fpcalc_download():
    """Download fpcalc and return path to archive file."""
    return download_file(fpcalc_download_url())


def fpcalc_extract(archive):
    """Extract archive with fpcalc executable."""
    if archive.endswith(".zip"):
        with zipfile.ZipFile(archive, "r") as zip_file:
            for member in zip_file.namelist():
                filename = os.path.basename(member)
                if filename == "fpcalc.exe":
                    source = zip_file.open(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)
    elif archive.endswith("tar.gz"):
        with tarfile.open(archive, "r:gz") as tar_file:
            for member in tar_file.getmembers():
                if member.isfile() and member.name.endswith("fpcalc"):
                    source = tar_file.extractfile(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)


def fpcalc_install():
    """Install fpcalc command line tool and return path to executable."""
    if fpcalc_is_installed():
        log.debug("Fpcalc is already installed.")
        return fpcalc_bin()
    log.critical("installing fpcalc")
    archive_path = fpcalc_download()
    fpcalc_extract(archive_path)
    st = os.stat(fpcalc_bin())
    os.chmod(fpcalc_bin(), st.st_mode | stat.S_IEXEC)
    assert fpcalc_is_installed()
    return fpcalc_bin()


def fpcalc_version_info():
    """Get fpcalc version"""
    try:
        r = subprocess.run([fpcalc_bin(), "-v"], stdout=subprocess.PIPE)
        return r.stdout.decode("utf-8").strip().split()[2]
    except FileNotFoundError:
        return "FPCALC not installed"


def ffprobe_bin() -> str:
    """Returns local path to ffprobe executable."""
    path = os.path.join(iscc.APP_DIR, "ffprobe-{}".format(FFPROBE_VERSION))
    if system() == "Windows":
        path += ".exe"
    return path


def ffmpeg_bin() -> str:
    """Returns local path to ffmpeg executable."""
    path = os.path.join(iscc.APP_DIR, "ffmpeg-{}".format(FFMPEG_VERSION))
    if system() == "Windows":
        path += ".exe"
    return path


def is_installed(fp: str) -> bool:
    """"Check if binary at `fp` exists and is executable"""
    return os.path.isfile(fp) and os.access(fp, os.X_OK)


def ffprobe_download_url():
    """Return system dependant download url."""
    urls = requests.get(FFPROBE_API_URL).json()
    return urls["bin"][system_tag()]["ffprobe"]


def ffmpeg_download_url():
    """Return system dependant download url."""
    urls = requests.get(FFPROBE_API_URL).json()
    return urls["bin"][system_tag()]["ffmpeg"]


def ffprobe_download():
    """Download ffprobe and return path to archive file."""
    md5 = FFPROBE_CHECKSUMS.get(system_tag())
    return download_file(ffprobe_download_url(), md5=md5)


def ffmpeg_download():
    """Download ffmpeg and return path to archive file."""
    md5 = FFMPEG_CHECKSUMS.get(system_tag())
    return download_file(ffmpeg_download_url(), md5=md5)


def ffprobe_extract(archive: str):
    """Extract ffprobe from archive."""
    fname = "ffprobe.exe" if system() == "Windows" else "ffprobe"
    with zipfile.ZipFile(archive) as zip_file:
        with zip_file.open(fname) as zf, open(ffprobe_bin(), "wb") as lf:
            shutil.copyfileobj(zf, lf)


def ffmpeg_extract(archive: str):
    """Extract ffprobe from archive."""
    fname = "ffmpeg.exe" if system() == "Windows" else "ffmpeg"
    with zipfile.ZipFile(archive) as zip_file:
        with zip_file.open(fname) as zf, open(ffmpeg_bin(), "wb") as lf:
            shutil.copyfileobj(zf, lf)


def ffprobe_install():
    """Install ffprobe command line tool and return path to executable."""
    if is_installed(ffprobe_bin()):
        log.debug("ffprobe is already installed")
        return ffprobe_bin()
    log.critical("installing ffprobe")
    archive_path = ffprobe_download()
    ffprobe_extract(archive_path)
    st = os.stat(ffprobe_bin())
    os.chmod(ffprobe_bin(), st.st_mode | stat.S_IEXEC)
    assert is_installed(ffprobe_bin())
    return ffprobe_bin()


def ffmpeg_install():
    """Install ffmpeg command line tool and return path to executable."""
    if is_installed(ffmpeg_bin()):
        log.debug("ffmpeg is already installed")
        return ffmpeg_bin()
    log.critical("installing ffmpeg")
    archive_path = ffmpeg_download()
    ffmpeg_extract(archive_path)
    st = os.stat(ffmpeg_bin())
    os.chmod(ffmpeg_bin(), st.st_mode | stat.S_IEXEC)
    assert is_installed(ffmpeg_bin())
    return ffmpeg_bin()


def ffprobe_version_info():
    """Get ffprobe version"""
    try:
        r = subprocess.run([ffprobe_bin(), "-version"], stdout=subprocess.PIPE)
        return (
            r.stdout.decode("utf-8")
            .strip()
            .splitlines()[0]
            .split()[2]
            .rstrip("-static")
            .rstrip("-tessu")
        )
    except FileNotFoundError:
        return "ffprobe not installed"


def ffmpeg_version_info():
    """Get ffmpeg version"""
    try:
        r = subprocess.run([ffmpeg_bin(), "-version"], stdout=subprocess.PIPE)
        return (
            r.stdout.decode("utf-8")
            .strip()
            .splitlines()[0]
            .split()[2]
            .rstrip("-static")
            .rstrip("-tessu")
        )
    except FileNotFoundError:
        return "ffmpeg not installed"


########################################################################################
# Java                                                                                 #
########################################################################################


def java_bin():
    java_path = shutil.which("java")
    if not java_path:
        java_path = java_custom_path()
    return java_path


def java_custom_path():
    if platform.system() == "Windows":
        java_path = os.path.join(iscc.APP_DIR, "jdk-16.0.2+7-jre", "bin", "java.exe")
    else:
        java_path = os.path.join(iscc.APP_DIR, "jdk-16.0.2+7-jre", "bin", "java")
    return java_path


def java_is_installed():
    return bool(shutil.which("java")) or is_installed(java_custom_path())


def java_install():
    if java_is_installed():
        log.debug("java already installed")
        return java_bin()
    log.critical("installing java")
    return jdk.install("16", impl="openj9", jre=True, path=iscc.APP_DIR)


def java_version_info():
    try:
        r = subprocess.run([java_bin(), "--version"], stdout=subprocess.PIPE)
        return r.stdout.decode("utf-8")
    except FileNotFoundError:
        return "JAVA not installed"


########################################################################################
# Apache Tika                                                                          #
########################################################################################


def tika_bin():
    # type: () -> str
    """Returns path to java tika app call"""
    return os.path.join(iscc.APP_DIR, f"tika-app-{TIKA_VERSION}.jar")


def tika_is_installed():
    # type: () -> bool
    """Check if tika is installed"""
    return os.path.exists(tika_bin())


def tika_download_url():
    # type: () -> str
    """Return tika download url"""
    return TIKA_URL


def tika_download():
    # type: () -> str
    """Download tika-app.jar and return local path"""
    return download_file(tika_download_url(), md5=TIKA_CHECKSUM)


def tika_install():
    # type: () -> str
    """Install tika-app.jar if not installed yet."""
    # Ensure JAVA is installed
    java_install()

    if tika_is_installed():
        log.debug("Tika is already installed")
        return tika_bin()
    else:
        log.critical("installing tika")
        path = tika_download()
        st = os.stat(tika_bin())
        os.chmod(tika_bin(), st.st_mode | stat.S_IEXEC)
        return path


def tika_version_info():
    # type: () -> str
    """
    Check tika-app version

    :return: Tika version info string
    :rtype: str
    """
    try:
        r = subprocess.run(
            [java_bin(), "-jar", tika_bin(), "--version"], stdout=subprocess.PIPE
        )
        return r.stdout.decode("utf-8")
    except FileNotFoundError:
        return "Tika not installed"


def system_tag():
    os_tag = system().lower()
    if os_tag == "darwin":
        os_tag = "osx"
    os_bits = architecture()[0].rstrip("bit")
    return f"{os_tag}-{os_bits}"


if __name__ == "__main__":
    print(java_version_info())
    print(tika_version_info())
    # print(tika_is_installed())
    # print(tika_install())
    # print(tika_is_installed())
    # print(tika_version_info())



# -*- coding: utf-8 -*-
"""Persistent inverted index for ISCC components and granular features.

ISCCs are stored in their raw-bytes representation. All other keys and values are
serialized to bytes with MessagePack.

We use LMDB `dupsort` to store multiple values per key and `set_range` to iterate
compatible (matchable) components by prefix. Features are indexed in separate
sub-databases per feature-type and version.


Database Structure:

    isccs:
        key -> iscc

    components:
        component -> key1..keyx

    features-<type>-<version>:
        feature .>  (key1, pos1)..(key1, pos1)

    metadata:
        key -> metadata

The `components` inverted index can be rebuild from the `isccs`-table.
The `features` inverted indexes can be rebuild from the `metadata`-table if available.
"""
import os
from os.path import join
from pathlib import Path
from typing import List, Optional, Dict, Any, Tuple, Union, Set, Generator
from bitarray import bitarray
from loguru import logger as log
import lmdb
import shutil
from humanize import naturalsize
from iscc.schema import FeatureMatch, Features, IsccMatch, QueryResult, ISCC
from iscc.options import SdkOptions
import msgpack
from annoy import AnnoyIndex
import iscc_core
from iscc.wrappers import decompose
from iscc_core.codec import Code, decode_base64, encode_base64, MT
from iscc.metrics import compare, distance_ba, distance_bytes


IsccObj = Union[str, Dict, Code, ISCC]
Key = Union[int, str, bytes]


class Index:
    def __init__(self, name="iscc-db", readonly=False, **options):
        # type: (str, bool, **Any) -> Index
        """Open a named index or create a new one with `name`.

        A given index can only have one concurrent writer. Multiple readers are
        allowed (open with readonly=True).

        :param str name: Name of the index (default iscc-db).
        :param bool readonly: Open index in readonly mode.
        :key str index_root: The root path for index databases (default APP_DIR).
        :key bool index_components: Create inverted index of components (default True).
        :key bool index_features: Create inverted index of features (default False).
        :key bool index_metadata: Store metadata in index (default False).
        """

        self.opts = SdkOptions(**options)

        self.name = name
        self.dbpath = join(self.opts.index_root, self.name)
        log.debug(f"init index storage at {self.dbpath}")
        os.makedirs(self.dbpath, exist_ok=True)
        self.env = lmdb.open(
            path=self.dbpath,
            map_size=2 ** 20,
            readonly=readonly,
            max_dbs=24,
            metasync=False,
            sync=False,
            readahead=False,
            writemap=False,
            meminit=False,
            map_async=True,
        )
        # Aproximate Nearest Neighbor indexes
        self.anns: Dict[str, AnnoyIndex] = {}

    def add(self, iscc_obj, key=None):
        # type: (IsccObj, Optional[Key]) -> Key
        """Add an ISCC to the index.

        The ISCC can be provided as a string or Code object. Alternatively you can pass
        the ISCC wrapped in a schema.ISCC object or conforming dict. This is required
        if granular features should be indexed or if you want to store ISCC metadata
        in the index.

        Optionally you may provide a unique key (str or int) to map entries to your
        external database. If no key is provided the index will create and return
        an autoincremented integer id. Do not mix both aproaches.

        :param IsccObj iscc_obj: ISCC str, Code, schema.ISCC or conforming dict.
        :param Key key: Optional custom primary key (int or str).
        :returns: Primary key of added (or pre-existing) iscc_obj.
        """

        iscc_code, features, metadata = self._parse_iscc_obj(iscc_obj)

        # Normalize
        components = decompose(iscc_code)
        iscc_code = iscc_core.gen_iscc_code_v0([c.code for c in components]).code_obj

        # Init component databases
        for c in components:
            self._db_components(c.type_id)

        # Check for duplicate ISCC
        exists = self.get_key(iscc_code)
        if exists is not None:
            return exists

        # Add canonical ISCC to main index
        db = self._db_isccs()
        if key is None:
            key = self._next_key()

        keyb = msgpack.dumps(key)

        self._put(db, keyb, iscc_code.bytes)

        # Add components to components index
        if self.opts.index_components:
            for code in components:
                self._add_component(code, keyb)

        # Add feature hashes
        if self.opts.index_features and features is not None:
            for fdict in features:
                pos = 0
                fobj = Features(**fdict)
                items = []
                sizes = fobj.sizes or list(range(len(fobj.features)))
                for feat, size in zip(fobj.features, sizes):
                    # feature -> (fkey, pos)
                    item = decode_base64(feat), msgpack.packb((keyb, pos))
                    items.append(item)
                    pos += size
                db = self._db_features(fobj.type_id)
                self._putmulti(db, items, dupdata=True, overwrite=True)

        # Add metadata
        if self.opts.index_metadata and metadata is not None:
            db = self._db_metadata()
            self._put(db, keyb, msgpack.dumps(metadata))

        return key

    def query(self, iscc_obj, k=10, ct=8, ft=4):
        # type: (IsccObj, int, int, int) -> QueryResult
        """Find nearest neighbours for ISCCs.

        We first match ISCCs by per-component similarity with a threshold of
        of maximum `ct` bits of hamming distance. IsccMatch results are sorted
        by distance (lowest distance first) and the top-`k` results are returned.

        Additionally granular features are also matched with a maximum threshold
        of `ft` bits if features are indexed and included in the query-`IsccObj`.
        The top-`k` `FeatureMatch` results are sorted by distance and will not
        include ISCCs that have already been matched by the previous component
        matching.

        :param IsccObj iscc_obj: ISCC string or ISCC metadata to be matched.
        :param int k: Return maximum top-k IsccMatches and top-k FeatureMatches.
        :param int ct: Match components with max hamming distance `ct`.
        :param int ft: Match features with max hamming distance `ft`
        :return: QueryResult
        """

        source_iscc, features, metadata = self._parse_iscc_obj(iscc_obj)
        iscc_matches = set()
        seen_fkeys = set()

        # Collect ISCC component level matches
        for comp_obj in reversed(decompose(source_iscc)):
            for fkey in self.match_component(comp_obj, ct=ct):
                if fkey not in seen_fkeys:
                    iscc_matches.add(self._build_match(source_iscc, fkey))
                    seen_fkeys.add(fkey)

        # Collect ISCC feature level matches
        feature_matches = set()
        for feat_dict in features:
            feat_obj = Features(**feat_dict)
            src_pos = 0
            sizes = feat_obj.sizes or list(range(len(features)))
            for src_feat, src_size in zip(feat_obj.features, sizes):
                for fm in self.match_feature(
                    feat_obj.type_id, src_feat, src_pos, ft=ft, ignore=seen_fkeys
                ):
                    feature_matches.add(fm)
                src_pos += src_size

        return QueryResult(
            iscc_matches=sorted(iscc_matches)[:k],
            feature_matches=sorted(feature_matches)[:k],
        )

    def match_component(self, code, ct=0):
        # type: (Code, int) -> List[bytes]
        """
        Match ISCCs (fkeys) by given commponent with threshold `ct` (max distance).

        :param Code code: ISCC Code component object to be matched.
        :param int ct: Component threshold (max hamming distance for match).
        :return: List[bytes] - foreign keys to ISCC table.
        """
        db = self._db_components(code.type_id)

        if code.maintype == MT.INSTANCE or ct == 0:
            # Simple get if instance code or threshold is 0
            return self._get_values(db, code.hash_bytes)
        else:
            # TODO: pluck ANNS search here if available
            # Scan for nearest neighbors
            fkeys = set()

            prefix = code.header_bytes
            with self.env.begin(db) as txn:
                with txn.cursor(db) as c:
                    for k in c.iternext_nodup(keys=True, values=False):
                        candidate = Code(prefix + k)
                        distance = distance_ba(code.hash_ba, candidate.hash_ba)
                        if distance > ct:
                            continue
                        c2 = txn.cursor(db)
                        c2.set_key(k)
                        for v in c2.iternext_dup(keys=False, values=True):
                            fkeys.add(v)
                        c2.close()
            return list(fkeys)

    def match_feature(self, kind, feature, src_pos=None, ft=0, ignore=None):
        # type: (str, Union[str,bytes], Optional[float], int, Set[bytes]) -> List[FeatureMatch]
        """
        Match ISCCs by a given feature with feature threshold `ft` (max distance).
        Ignore matches for fkeys in `ignore`.

        :param str kind: Feature type to be matched.
        :param str|bytes feature: Raw or base64 encoded granular feature.
        :param float|None src_pos: Position of queried feature in source.
        :param int ft: Feature threshold (max hamming distance to match feature)
        :param set ignore: Set of ISCC table primarky keys to ignore.
        """

        if isinstance(feature, str):
            feature_raw, feature_str = decode_base64(feature), feature
        elif isinstance(feature, bytes):
            feature_raw, feature_str = feature, encode_base64(feature)
        else:
            raise ValueError(f"feature must be str or bytes, not {type(feature)}")

        ignore = ignore or set()

        matches = set()

        db = self._db_features(kind)

        if ft == 0:
            # Simple lookup with zero threshold
            lookup = self._get_feature(kind, feature_raw)
            for fkey, match_pos in lookup:
                if fkey in ignore:
                    continue
                fm = FeatureMatch(
                    key=msgpack.unpackb(fkey),
                    matched_iscc=self.get_iscc(fkey).code,
                    kind=kind,
                    source_feature=feature_str,
                    source_pos=src_pos,
                    matched_feature=feature_str,
                    matched_position=match_pos,
                )
                matches.add(fm)
        elif self.anns.get("feat-" + kind):
            # ANNS based lookup
            anns: AnnoyIndex = self.anns.get("feat-" + kind)
            feature_ba = bitarray()
            feature_ba.frombytes(feature_raw)
            result = anns.get_nns_by_vector(
                feature_ba.tolist(True), n=10, search_k=-1, include_distances=True
            )
            # filter by distance, dereference annoy_id and encode vectors to hashes
            mfeatures_raw, mfeatures_b64, distances = [], [], []
            for annoy_id, dist in zip(result[0], result[1]):
                if dist <= ft:
                    mfraw = bitarray(anns.get_item_vector(annoy_id)).tobytes()
                    mfeatures_raw.append(mfraw)
                    mfeatures_b64.append(encode_base64(mfraw))
                    distances.append(int(dist))
            # lookup matchesd features in lmdb
            with self.env.begin() as txn:
                with txn.cursor(db=db) as c:
                    refs = c.getmulti(mfeatures_raw, dupdata=True)
                # decode refs > [(mfeature_raw, msgpack_coded(fkey, pos)), ...]
                isccs_raw, fkeys, positions = [], [], []
                for ref in refs:
                    fkey, pos = msgpack.unpackb(ref[1])
                    fkeys.append(fkey)
                    positions.append(pos)
                # deference and decode matched ISCCs
                with txn.cursor(self._db_isccs()) as c:
                    isccs = [Code(v).code for _, v in c.getmulti(fkeys)]
            # Build FeatureMatch objects
            # iscc str, kind, src_feat, src_pos, mat_feat, mat_pos
            for iscc_str, matched_feat, matched_pos, dist, fkey in zip(
                isccs, mfeatures_b64, positions, distances, fkeys
            ):
                if fkey in ignore:
                    continue
                fm = FeatureMatch(
                    key=msgpack.unpackb(fkey),
                    matched_iscc=iscc_str,
                    kind=kind,
                    source_feature=feature_str,
                    source_pos=src_pos,
                    matched_feature=matched_feat,
                    matched_position=matched_pos,
                    distance=dist,
                )
                matches.add(fm)
        else:
            # Fallback full scan with non-zero threshold an no ANNS index
            db = self._db_features(kind)
            with self.env.begin(db) as txn:
                with txn.cursor(db) as c:
                    while c.next_nodup():
                        candidate_feature = c.key()
                        distance = distance_bytes(feature_raw, candidate_feature)
                        if distance > ft:
                            continue
                        for value in c.iternext_dup(keys=False, values=True):
                            fkey, matched_position = msgpack.loads(value)
                            if fkey in ignore:
                                continue
                            fm = FeatureMatch(
                                key=msgpack.unpackb(fkey),
                                matched_iscc=self.get_iscc(fkey).code,
                                kind=kind,
                                source_feature=feature_str,
                                source_pos=src_pos,
                                matched_feature=encode_base64(candidate_feature),
                                matched_position=matched_position,
                            )
                            matches.add(fm)

        return sorted(matches)

    def get_key(self, code: Code) -> Optional[int]:
        """Resolve internal ISCC key via components."""
        # Find per component matches
        components = decompose(code)
        idxs = []

        with self.env.begin() as txn:
            for comp_code in components:
                db = self._db_components(comp_code.type_id)
                # TODO fails with InvalidParameterError after DB reopen
                idx = txn.get(comp_code.hash_bytes, db=db)
                if idx is not None:
                    idxs.append(idx)
            # Check if any of the full code entries is an exact match
            full_code_bytes = iscc_core.gen_iscc_code_v0(
                [c.code for c in components]
            ).code_obj.bytes
            db = self._db_isccs()
            for idx in idxs:
                if txn.get(idx, db=db) == full_code_bytes:
                    return msgpack.loads(idx)

    def get_iscc(self, key):
        # type: (Key) -> Code
        """Get ISCC by index key"""
        if not isinstance(key, bytes):
            key = msgpack.dumps(key)
        db = self._db_isccs()
        with self.env.begin(db) as txn:
            iscc_bytes = txn.get(key)
            if iscc_bytes:
                return Code(iscc_bytes)

    def get_metadata(self, key):
        # type: (Key) -> dict
        """Get ISCC metadata by index key"""
        if not isinstance(key, bytes):
            key = msgpack.dumps(key)
        db = self._db_metadata()
        with self.env.begin(db) as txn:
            iscc_bytes = txn.get(key)
            if iscc_bytes:
                return msgpack.loads(iscc_bytes)

    def iter_isccs(self) -> Generator[bytes, None, None]:
        """Iterates over all indexed ISCC codes in insertion order."""
        db = self._db_isccs()
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                while c.next():
                    yield c.value()

    def iter_components(self) -> Generator[bytes, None, None]:
        """Itereates over all indexed components in lexicographic order."""
        with self.env.begin() as txn:
            for dbname in self.dbs():
                if dbname.startswith(b"comp-"):
                    db = self._db_components(dbname.lstrip(b"comp-").decode("ascii"))
                    with txn.cursor(db) as c:
                        for key in c.iternext_nodup():
                            yield key

    def dbs(self) -> List[bytes]:
        """Return a list of existing sub-databases in the main index."""
        dbnames = []
        with self.env.begin() as txn:
            c = txn.cursor()
            while c.next():
                dbnames.append(c.key())
        return dbnames

    def build_anns(self):
        """Build and reload ANNS indexes (replacing previous versions)."""
        tmp_files = []
        with self.env.begin() as txn:
            for dbname in self.dbs():
                # Filter relevant dbnames
                if b"INSTANCE" in dbname or not b"-" in dbname:
                    continue
                prefix = dbname.split(b"-")[0]
                kind = dbname.lstrip(prefix + b"-").decode()
                db = (
                    self._db_components(kind)
                    if prefix == b"comp"
                    else self._db_features(kind)
                )
                ann = AnnoyIndex(64, "hamming")
                log.debug(f"collect {kind} for anns")
                with txn.cursor(db) as c:
                    for i, k in enumerate(c.iternext_nodup(keys=True, values=False)):
                        ba = bitarray()
                        ba.frombytes(k)
                        ann.add_item(i, ba.tolist(True))
                log.debug(f"build {kind}")
                ann.build(-1)
                outf = join(self.dbpath, dbname.decode("ascii") + ".ann.tmp")
                log.debug(f"save {outf}")
                ann.save(outf)
                ann.unload()
                tmp_files.append(outf)
        # Switch to new anns files
        self.unload_anns()
        targets = [p.rstrip(".tmp") for p in tmp_files]
        for src, dst in zip(tmp_files, targets):
            try:
                os.remove(dst)
            except FileNotFoundError:
                pass
            os.rename(src, dst)
        self.load_anns()

    def load_anns(self):
        """Load existing ANNS indexes"""
        for ann_file in Path(self.dbpath).glob("*.ann"):
            ann = AnnoyIndex(64, "hamming")
            ann.load(ann_file.as_posix())
            self.anns[ann_file.stem] = ann

    def unload_anns(self):
        """Unload ANNS"""
        for ann in self.anns.values():
            ann.unload()
        self.anns = {}

    def destroy(self):
        """Close and delete index from disk."""
        self.env.close()
        log.debug(f"delete index storage at {self.dbpath}")
        shutil.rmtree(self.dbpath)

    def close(self):
        """Close index."""
        self.env.close()

    @property
    def map_size(self) -> int:
        return self.env.info()["map_size"]

    @property
    def stats(self):
        result = {}
        dbs = self.dbs()
        for db_name in dbs:
            db = self.env.open_db(db_name)
            with self.env.begin() as txn:
                entries = txn.stat(db)["entries"]
                result[db_name.decode("ascii")] = entries
        return result

    def _db_isccs(self) -> lmdb._Database:
        return self.env.open_db(b"isccs", integerkey=False, create=True)

    def _db_components(self, type_id: str) -> lmdb._Database:
        """Return componets database."""
        return self.env.open_db(
            b"comp-" + type_id.encode("ascii"),
            dupsort=True,  # Duplicate keys allowed
            create=True,  # Create table if required
            integerkey=False,  # Keys are component raw bytes
            integerdup=False,  # Values are raw byte foreign keys to iscc table
            dupfixed=False,  # Variable length values
        )

    def _db_features(self, kind: str) -> lmdb._Database:
        return self.env.open_db(
            b"feat-" + kind.encode("utf-8"),
            dupsort=True,  # Duplicate keys allowed
            create=True,  # Create table if required
            integerkey=False,  # Keys are raw byte feature hashes
            integerdup=False,  # Values are msgpack serialized tuples (fkey, size)
            dupfixed=False,  # Variable length values
        )

    def _db_metadata(self) -> lmdb._Database:
        return self.env.open_db(b"metadata", integerkey=False, create=True)

    def _add_component(self, code: Code, fkey: bytes) -> bool:
        """
        Add a component to the index with a pointer to its source ISCC.
        A single component can point to multipe ISCC entries in the main table.
        """
        db = self._db_components(code.type_id)
        return self._put(db, code.hash_bytes, fkey, dupdata=True, overwrite=True)

    def _add_feature(self, kind, feature, fkey, position):
        # type: (str, bytes, bytes, Union[int, float]) -> bool
        """Add a feature to the index with a pointer to its source ISCC and position.
        A single feature can point to multiple ISCC entries at multiple positions.
        """
        db = self._db_features(kind)
        value = msgpack.packb((fkey, position))
        return self._put(db, feature, value, dupdata=True, overwrite=True)

    def _get_values(self, db, key):
        # type: (lmdb._Database, bytes) -> List[bytes]
        """Collect all values for a given `key` from LMDB dupsort database."""
        values = []
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                value = c.get(key)
                if value is None:
                    return []
                values.append(value)
                while c.next_dup():
                    values.append(c.value())
        return values

    def _get_feature(self, kind, feature):
        # type: (str, bytes) -> List[Tuple[bytes, Union[int, float]]]
        """Get a list of (fkey, position) results for a given feature"""
        db = self._db_features(kind)
        results = []
        with self.env.begin() as txn:
            with txn.cursor(db) as c:
                r = c.get(feature)
                if r is None:
                    return []
                results.append(msgpack.unpackb(r, use_list=False))
                while c.next_dup():
                    results.append(msgpack.unpackb(c.value(), use_list=False))
        return results

    def _next_key(self) -> bytes:
        """Next free autoincrement key"""
        db = self._db_isccs()
        with self.env.begin(db) as txn:
            with txn.cursor(db) as c:
                empty = not c.last()
                key = msgpack.loads(c.key()) + 1 if not empty else 0
        return key

    def _put(self, db, key: bytes, value: bytes, dupdata=True, overwrite=True) -> bool:
        """Wrap LMDB put in a transaction and auto-resize db if required."""
        try:
            with self.env.begin(db, write=True) as txn:
                return txn.put(key, value, dupdata=dupdata, overwrite=overwrite)
        except lmdb.MapFullError:
            new_size = self.map_size * 2
            log.info(f"Resizing {self.dbpath} to {naturalsize(new_size)}")
            self.env.set_mapsize(self.map_size * 2)
            with self.env.begin(db, write=True) as txn:
                return txn.put(key, value, dupdata=dupdata, overwrite=overwrite)

    def _putmulti(self, db, items, dupdata=True, overwrite=True) -> Tuple[int, int]:
        """Wrap LMDB putmulti in a transaction and auto-resize db if required."""
        try:
            with self.env.begin(db, write=True) as txn:
                with txn.cursor(db) as c:
                    return c.putmulti(items, dupdata=dupdata, overwrite=overwrite)
        except lmdb.MapFullError:
            new_size = self.map_size * 2
            log.info(f"Resizing {self.dbpath} to {naturalsize(new_size)}")
            self.env.set_mapsize(self.map_size * 2)
            with self.env.begin(db, write=True) as txn:
                with txn.cursor(db) as c:
                    return c.putmulti(items, dupdata=dupdata, overwrite=overwrite)

    def _build_match(self, source_iscc, matched_key):
        # type: (Code, bytes) -> IsccMatch
        """Build IsccMatch for source_iscc and iscc of a matched foreign key."""
        matched_iscc: Code = self.get_iscc(matched_key)
        matchdata = compare(source_iscc, matched_iscc)
        matchdata["matched_iscc"] = matched_iscc.code
        matchdata["key"] = msgpack.loads(matched_key)
        matchdata["distance"] = distance_ba(source_iscc.hash_ba, matched_iscc.hash_ba)
        return IsccMatch(**matchdata)

    @staticmethod
    def _parse_iscc_obj(iscc_obj):
        # type: (IsccObj) -> Tuple[Code, List[dict], Optional[dict]]
        """Unpack different types of ISCC inputs."""

        metadata = None
        features = []

        if isinstance(iscc_obj, str):
            iscc_code = Code(iscc_obj)
        elif isinstance(iscc_obj, Code):
            iscc_code = iscc_obj
        elif isinstance(iscc_obj, ISCC):
            iscc_code = Code(iscc_obj.iscc)
            metadata = iscc_obj.dict(exclude_unset=True)
            features = metadata.get("features", [])
        elif isinstance(iscc_obj, dict):
            iscc_code = Code(iscc_obj["iscc"])
            metadata = iscc_obj
            features = metadata.get("features", [])
        else:
            raise ValueError(
                f"'iscc_obj' must be one of {IsccObj} not {type(iscc_obj)}."
            )
        return iscc_code, features, metadata

    def _dump(self):
        """Dump DB to console"""
        for idx, value in enumerate(self.iter_isccs()):
            print("ISCC", idx, Code(value).code)
        if self.opts.index_components:
            db = self._db_components()
            with self.env.begin(db) as txn:
                with txn.cursor(db) as c:
                    c.first()
                    for k in c.iternext_nodup():
                        print("COMPONENT", Code(k), end=" ")
                        c2 = txn.cursor(db)
                        c2.set_key(k)
                        for v in c2.iternext_dup(keys=False, values=True):
                            print(msgpack.loads(v), end=" ")
                        print()
        if self.opts.index_features:
            kinds = ["text", "image", "audio", "video", "data"]
            for kind in kinds:
                try:
                    db = self._db_features(kind)
                except lmdb.ReadonlyError:
                    continue
                with self.env.begin(db) as txn:
                    with txn.cursor(db) as c:
                        c.first()
                        for k in c.iternext_nodup():
                            print(kind.upper(), end=" ")
                            print(encode_base64(k), end=" ")
                            c2 = txn.cursor(db)
                            c2.set_key(k)
                            for v in c2.iternext_dup(keys=False, values=True):
                                fkey, pos = msgpack.loads(v)
                                value = [msgpack.loads(fkey), round(pos, 3)]
                                print(value, end=" ")
                            print()
        if self.opts.index_metadata:
            db = self._db_metadata()
            with self.env.begin(db) as txn:
                for k, v in txn.cursor(db):
                    print("METADATA", msgpack.loads(k), msgpack.loads(v))

    def __len__(self):
        """Number of indexed ISCCs"""
        db = self._db_isccs()
        with self.env.begin(db) as txn:
            stat = txn.stat(db)
            return stat.get("entries")

    def __contains__(self, item):
        """Check if full iscc code is in index."""
        key = self.get_key(item)
        return False if key is None else True



# -*- coding: utf-8 -*-
"""ISCC Reference Implementation"""
import gzip
import os.path
import iscc_core
from os.path import basename
import base64
import io
import mobi
from codetiming import Timer
from humanize import naturalsize
from iscc_core.code_data import DataHasherV0
from iscc_core.codec import AnyISCC
from loguru import logger as log
from typing import List, Union, Any
from PIL import Image
from blake3 import blake3
import iscc
from iscc import jcs
from iscc.mp7 import read_ffmpeg_signature
from iscc.schema import (
    GMT,
    Uri,
    Data,
    File,
    Readable,
    ISCC,
)
from iscc.options import SdkOptions, sdk_opts
from iscc.mediatype import mime_guess, mime_to_gmt
from iscc import uread
from iscc.wrappers import decompose
from iscc_core import codec


###############################################################################
# High-Level ISCC Code generator functions                                   #
###############################################################################


def code_iscc(uri, title=None, extra=None, **options):
    # type: (Union[Uri, File], Union[str, bytes], Union[None, str, bytes, dict], **Any) -> dict
    """Create a full ISCC.

    The full ISCC is a composite of Meta, Content, Data and Instance Codes.

    :param uri: File or filepath used for ISCC creation.
    :param title: Title of media asset (defaults to extracted metadata or filename)
    :param extra: Metadata to be used for Meta-Code generation
    :param options: See iscc.schema.Options for detailed ISCC generator options.
    """
    opts = SdkOptions(**options)
    result = {"version": "0-0-0"}
    features = []

    file_obj = uread.open_data(uri)

    file_name = getattr(file_obj, "name", None)
    if file_name:
        result["filename"] = basename(file_name)

    with Timer(text="{:0.4f}s for instance code creation", logger=log.debug):
        instance = code_instance(file_obj, **options)
    result.update(instance)

    with Timer(text="{:0.4f}s for data code creation", logger=log.debug):
        data = code_data(file_obj, **options)

    if "features" in data:
        features.append(data.pop("features"))
    result.update(data)

    content = code_content(file_obj, **options)

    # TODO generalize to store extracts from all mediatypes (mp7, chroma,...)
    if opts.text_store and "plaintext" in content:
        plaintext = content.pop("plaintext")
        path = os.path.dirname(file_obj.name)
        outpath = os.path.join(path, instance["datahash"] + ".txt.gz")
        with gzip.open(outpath, "wb") as outf:
            outf.write(plaintext.encode("utf-8"))
        log.debug(f"Stored plaintext at: {outpath}")

    if "features" in content:
        features.append(content.pop("features"))
    result.update(content)

    if features:
        result["features"] = features

    if not title:
        title = content.get("title")
    if not title and file_name:
        title = iscc.text.title_from_uri(file_name)

    meta = code_meta(title, extra, **options)
    result.update(meta)

    iscc_code_obj = iscc_core.gen_iscc_code_v0(
        [meta["iscc"], content["iscc"], data["iscc"], instance["iscc"]]
    ).code_obj
    result["iscc"] = iscc_code_obj.code
    concat = bytes.fromhex(result["metahash"]) + bytes.fromhex(result["datahash"])
    result["tophash"] = blake3(concat).hexdigest()
    valid = ISCC(**result)

    return valid.dict(exclude_unset=True)


def code_meta(title, extra=None, **options):
    # type: (Union[str, bytes], Union[None, str, bytes, dict], **Any) -> dict
    """Generate Meta Code from title and extra metadata.

    :param title: Used as input for first half of Meta-Code (or full if no extra)
    :param extra: Extended (str or dict) metadata used as input for second half of code.
    :key meta_bits: Length of generated Meta-Code in bits
    :key meta_ngram_size: Number of characters for sliding window over metadata text.
    :key meta_trim_title: Trim title length to this mumber of bytes
    :key meta_trim_extra: Trim extra data to this number of bytes
    :returns: Dict keys: code, title, matahash, (extra)
    """
    opts = SdkOptions(**options)
    extra = None if extra in (b"", "", bytearray(), None) else extra

    if isinstance(extra, dict):
        extra = jcs.canonicalize(extra).decode("utf-8")
        log.debug("Metadata JSON encoded and conanicalized with JCS")

    meta_code = iscc_core.gen_meta_code_v0(
        title=title, extra=extra, bits=opts.meta_bits
    )
    return meta_code.dict(exclude_unset=True, exclude={"binary"}, exclude_none=True)


def code_content(data, **options):
    # type: (Union[Uri, File], **Any) -> dict
    """Detect mediatype and create corresponding Content-Code."""
    mediatype = mime_guess(data)
    gmt = mime_to_gmt(mediatype)

    is_mobi = mediatype == "application/x-mobipocket-ebook"
    if is_mobi:
        tempdir, data = mobi.extract(data)

    if gmt == GMT.text:
        with Timer(text="{:0.4f}s for content code text creation", logger=log.debug):
            cc = code_text(data, **options)
    elif gmt == GMT.image:
        with Timer(text="{:0.4f}s for content code image creation", logger=log.debug):
            cc = code_image(data, **options)
    elif gmt == GMT.audio:
        with Timer(text="{:0.4f}s for content code audio creation", logger=log.debug):
            cc = code_audio(data, **options)
    elif gmt == GMT.video:
        with Timer(text="{:0.4f}s for content code video creation", logger=log.debug):
            cc = code_video(data, **options)
    else:
        raise ValueError("Unsupported mediatype: {}".format(mediatype))

    cc["mediatype"] = mediatype
    cc["gmt"] = gmt

    return cc


def code_text(data, **options):
    # type: (Readable, **Any) -> dict
    """Generate Content-Code Text

    :param data: Any kind of text document
    """
    opts = SdkOptions(**options) if options else sdk_opts
    result = {}

    # f = uread.open_data(data)
    text_raw = iscc.text.extract_text(data)
    metadata = iscc.text.extract_text_metadata(data, text_raw, **options)
    result.update(metadata)

    text_norm = iscc.text.normalize_text(text_raw)
    digest_256 = iscc_core.code_content_text.soft_hash_text_v0(text_norm.lower())

    text_code = iscc_core.codec.encode_component(
        mtype=iscc_core.codec.MT.CONTENT,
        stype=iscc_core.codec.ST_CC.TEXT,
        version=codec.VS.V0,
        length=opts.text_bits,
        digest=digest_256,
    )
    result["iscc"] = text_code

    granular = (
        opts.all_granular if isinstance(opts.all_granular, bool) else opts.text_granular
    )
    if granular:
        features = iscc.text.extract_text_features(text_norm, **options)
        result["features"] = features

    if opts.text_store:
        result["plaintext"] = text_raw

    return result


def code_image(data, **options):
    # type: (Union[Readable], **Any) -> dict
    """Generate Content-Code Image

    :param data: raw data, file or path to image
    :key bool image_granular: Wether to compute granular image features
    :key int image_granular_n: Number of features to compute (default 32)
    """
    opts = SdkOptions(**options) if options else sdk_opts

    try:
        result = iscc.image.extract_image_metadata(data) or {}
    except Exception as e:
        log.error(f"Failed image metadata extraction: {e}")
        result = {}

    stream = uread.open_data(data)
    imo = Image.open(stream)
    pixels = iscc.image.normalize_image(imo, **options)
    image_code = iscc_core.gen_image_code_v0(pixels, bits=opts.image_bits)
    image_code.width, image_code.height = imo.size
    result.update(image_code.dict(exclude_unset=True))

    if isinstance(data, Image.Image):
        img_obj = data
    else:
        img_obj = Image.open(io.BytesIO(uread.open_data(data).read()))

    granular = (
        opts.all_granular
        if isinstance(opts.all_granular, bool)
        else opts.image_granular
    )
    if granular:
        try:
            feat_obj = iscc.image.extract_image_features(data, n=opts.image_granular_n)
            result["features"] = feat_obj
        except Exception as e:
            log.error("image feature extraction failed")
            log.exception(e)

    do_preview = (
        opts.all_preview if isinstance(opts.all_preview, bool) else opts.image_preview
    )

    if do_preview:
        preview = iscc.image.extract_image_preview(img_obj, **options)
        preview_uri = iscc.image.encode_image_to_data_uri(preview, **options)
        result["preview"] = preview_uri

    return result


def code_audio(data, **options):
    # type: (Union[Uri, Data, List], **Any) -> dict
    """Generate Audio-ID from file(path) or Chromaprint features"""
    opts = SdkOptions(**options)
    result = dict()

    if isinstance(data, list):
        chroma = dict(fingerprint=data)
    else:
        chroma = iscc.audio.extract_audio_features(data, **options)
        metadata = iscc.audio.extract_audio_metadata(data)
        result.update(metadata)

    audio_code = iscc_core.gen_audio_code_v0(
        cv=chroma["fingerprint"], bits=opts.audio_bits
    )
    granular = (
        opts.all_granular
        if isinstance(opts.all_granular, bool)
        else opts.audio_granular
    )
    if granular:
        features = iscc.audio.encode_audio_features(chroma["fingerprint"])
        result["features"] = features

    result["iscc"] = audio_code.iscc
    return result


def code_video(uri, **options):
    # type: (Union[Uri, File], **Any) -> dict
    """Compute Content-ID video."""
    opts = SdkOptions(**options)
    result = {}
    metadata = iscc.video.extract_video_metadata(uri)
    result.update(metadata)

    crop_value = iscc.video.detect_video_crop(uri) if opts.video_crop else None

    granular = (
        opts.all_granular
        if isinstance(opts.all_granular, bool)
        else opts.video_granular
    )
    do_ffmpeg_scenes = granular and opts.video_scenes_ffmpeg

    if do_ffmpeg_scenes:
        signature, cutpoints = iscc.video.extract_video_signature_cutpoints(
            uri, crop_value, **options
        )
    else:
        signature = iscc.video.extract_video_signature(uri, crop_value, **options)

    with Timer(text="{:0.4f}s for video signature decoding", logger=log.debug):
        frames = read_ffmpeg_signature(signature)
    log.debug(f"{naturalsize(len(signature))} video sig with {len(frames)} frames")
    features = [tuple(sig.vector.tolist()) for sig in frames]
    video_code = iscc_core.gen_video_code_v0(features, bits=opts.video_bits)
    result["iscc"] = video_code.iscc

    if opts.video_include_fingerprint:
        result["fingerprint"] = base64.b64encode(signature).decode("ascii")

    do_preview = (
        opts.all_preview if isinstance(opts.all_preview, bool) else opts.video_preview
    )

    if do_preview:
        img_raw = iscc.video.extract_video_preview(uri)
        result["preview"] = iscc.image.encode_image_to_data_uri(img_raw)

    if not granular:
        return result

    if opts.video_scenes:
        if not opts.video_scenes_ffmpeg:
            cutpoints = iscc.video.detect_video_scenes(uri, **options)
        features = iscc.video.compute_video_features_scenes(frames, cutpoints)
    else:
        features = iscc.video.compute_video_features_rolling(frames, **options)

    result["features"] = features

    return result


def code_data(data, **options):
    # type: (Readable, **Any) -> dict
    """Create ISCC Data-Code.

    The Data-Code is a similarity preserving hash of the input data.

    :param Readable data: File, filepath or raw data used for Data-Code creation.
    :key int data_avg_chunk_size: Target chunk size in bytes for data chunking.
    :key int data_granular: Return granular features (one hash per chunk).
    :key int data_granular_factor: Size of granular data chunks times average chunk size.
    :key int io_chunk_size: Number of bytes to read per IO operation.
    :return: DataCode conformant dictionary
    :rtype: dict
    """

    opts = SdkOptions(**options)
    stream = uread.open_data(data)

    hasher = DataHasherV0()
    data = stream.read(opts.io_read_size)

    while data:
        hasher.push(data)
        data = stream.read(opts.io_read_size)

    code = hasher.code(bits=opts.data_bits)
    result = dict(iscc=code)

    granular = (
        opts.all_granular if isinstance(opts.all_granular, bool) else opts.data_granular
    )

    if granular:
        features = iscc.data.encode_data_features(
            hasher.chunk_sizes, hasher.chunk_features
        )
        result["features"] = features

    return result


def code_instance(data, **options):
    # type: (Readable, **Any) -> dict
    """Create ISCC Instance-Code.

    The Instance-Code is prefix of a cryptographic hash (blake3) of the input data.
    It´s purpose is to serve as an checksum that detects even minimal changes
    to the data of the referenced media asset. For cryptographicaly secure integrity
    checking a full 256-bit blake3 hash is provided with the `datahash` field.

    :param Readable data: File, filepath or raw data used for Instance-Code creation.
    :key int instance_bits: Length of generated Instance-Code in bits (default 64).
    :key int io_chunk_size: Number of bytes to read per IO operation.
    :return: An InstanceCode conformant dict with attributes: code, datahash, filesize
    :rtype: dict
    """
    opts = SdkOptions(**options)
    stream = uread.open_data(data)
    code = iscc_core.gen_instance_code_v0(stream, opts.instance_bits)
    return code.dict()


def code_iscc_id(chain_id, iscc_code, uc=0):
    # type: (Union[int,codec.ST_ID], AnyISCC, int) -> dict
    """Create ISCC-ID.

    :param int chain_id: ID of chain where the ISCC registration happend.
    :param AnyISCC iscc_code: An ISCC.
    :param in uc: Uniqueness counter.
    """

    # Normalize
    components = decompose(iscc_code)
    iscc_code = iscc_core.gen_iscc_code_v0([c.code for c in components])

    return iscc_core.gen_iscc_id_v0(chain_id, iscc_code.iscc, uc).dict()



# -*- coding: utf-8 -*-
__version__ = "1.1.0b14"
import os
from pathlib import Path
import click

HERE = Path(__file__).parent
APP_NAME = "iscc"
APP_DIR = click.get_app_dir(APP_NAME, roaming=False)
os.makedirs(APP_DIR, exist_ok=True)
os.environ["LOGURU_AUTOINIT"] = "False"


from iscc.api import (
    code_iscc,
    code_meta,
    code_content,
    code_text,
    code_image,
    code_audio,
    code_video,
    code_data,
    code_instance,
    code_iscc_id,
)

from iscc import (
    text,
    image,
    audio,
    video,
    match,
    metrics,
    mediatype,
    utils,
    bin,
    data,
    index,
    jcs,
    uread,
    utils,
)



# -*- coding: utf-8 -*-
"""Content-ID Image functions"""
import base64
from collections import defaultdict
import io
import pyexiv2
from iscc_core.codec import encode_base64
from loguru import logger
from pathlib import Path
from typing import Any, Optional, Sequence, Union
from PIL import Image, ImageChops, ImageEnhance, ImageOps
import numpy as np
from more_itertools import chunked
from iscc.schema import FeatureType, Readable
from iscc_core.simhash import similarity_hash
from iscc.text import normalize_text
from iscc.options import SdkOptions, sdk_opts
from iscc import uread


IMAGE_META_MAP = {
    "Xmp.dc.title": "title",
    "Xmp.dc.identifier": "identifier",
    "Xmp.dc.language": "language",
    "Xmp.xmp.Identifier": "identifier",
    "Xmp.xmp.Nickname": "title",
    "Xmp.xmpDM.shotName": "title",
    "Xmp.photoshop.Headline": "title",
    "Xmp.iptcExt.AOTitle": "title",
    "Iptc.Application2.Headline": "title",
    "Iptc.Application2.ObjectName": "title",
    "Iptc.Application2.Language": "language",
    "Exif.Image.ImageID": "identifier",
    "Exif.Image.XPTitle": "title",
    "Exif.Photo.ImageUniqueID": "identifier",
}


def extract_image_metadata(data):
    # type: (Readable) -> Optional[dict]
    """Extracts and normalizes seed metadata from image files (exif, iptc, xmp).

    TODO: Create a more advanced cross-standard metadata mapping.
    Returns an optional dictionary with the following possible fields:
        - title
        - identifier
        - langauge
    """
    file = uread.open_data(data)
    try:
        img_obj = pyexiv2.ImageData(file.read())
        meta = {}
        meta.update(img_obj.read_exif())
        meta.update(img_obj.read_iptc())
        meta.update(img_obj.read_xmp())
    except Exception as e:
        logger.warning(f"Image metadata extraction failed: {e}")
        return None

    selected_meta = defaultdict(set)
    for k, v in meta.items():
        if k not in IMAGE_META_MAP:
            continue
        if isinstance(v, list):
            v = tuple(normalize_text(item) for item in v)
            if not v:
                continue
            v = ";".join(v)
            v = v.replace('lang="xdefault"', "").strip()
        elif isinstance(v, str):
            v = normalize_text(v)
            v = v.replace('lang="xdefault"', "").strip()
            if not v:
                continue
        elif isinstance(v, dict):
            v = v.get('lang="x-default"', None)
        else:
            raise ValueError(f"missed type {type(v)}")

        field = IMAGE_META_MAP[k]
        selected_meta[field].add(v)

    longest_meta = {}
    for k, v in selected_meta.items():
        # value is a set of candidates
        best_v = max(selected_meta[k], key=len)
        if not best_v:
            continue
        longest_meta[k] = best_v
    return longest_meta or None


def extract_image_features(data, n=32):
    # type: (Readable, int) -> dict
    """Extract granular features from image.

    Returns a dict (features, positions, sizes) where:
        features - base64 encoded feature hashes
        sizes - keypoint sizes as percentage relative to the larger side of the image
        positions - percentage based x, y coordinates of the corresponding keypoints
    """
    try:
        import cv2 as cv
    except ImportError as e:
        raise EnvironmentError(
            "Please install opencv-contrib-python-headless for granular image support"
        )

    file = uread.open_data(data)
    img_array = np.array(bytearray(file.read()), dtype=np.uint8)
    img = cv.imdecode(img_array, cv.IMREAD_GRAYSCALE)

    height, width = img.shape[:2]
    pix_count = max(img.shape[:2])  # pix count of longer side
    orb = cv.ORB_create(32)
    keypoints = orb.detect(img, None)
    beblid = cv.xfeatures2d.BEBLID_create(0.75)
    keypoints, descriptors = beblid.compute(img, keypoints)
    zipped = list(zip(keypoints, descriptors))
    ranked = sorted(zipped, key=lambda x: x[0].response, reverse=True)
    feat_map = {}  # simhash -> (size, pos)
    for kp, desc in ranked:
        shash = encode_base64(similarity_hash([bytes(c) for c in chunked(desc, 8)]))
        feature_size = round(kp.size / pix_count * 100, 3)
        feature_pos = (
            round(kp.pt[0] / width * 100, 3),
            round(kp.pt[1] / height * 100, 3),
        )
        feat_map[shash] = (feature_size, feature_pos)
        if len(feat_map) == n:
            break
    features = list(feat_map.keys())
    sizes = [fm[0] for fm in feat_map.values()]
    positions = [fm[1] for fm in feat_map.values()]
    return dict(
        kind=FeatureType.image.value,
        version=0,
        features=features,
        sizes=sizes,
        positions=positions,
    )


def extract_image_preview(img, **options):
    # type: (Union[str, Path, Image.Image], **Any) -> Image.Image
    """Create a thumbnail for an image."""
    opts = SdkOptions(**options) if options else sdk_opts
    size = opts.image_preview_size
    if not isinstance(img, Image.Image):
        img = Image.open(img)
    else:
        # Pillow thumbnail operates inplace - avoid side effects.
        img = img.copy()
    img.thumbnail((size, size), resample=Image.LANCZOS)
    return ImageEnhance.Sharpness(img.convert("RGB")).enhance(1.4)


def encode_image_to_data_uri(img, **options):
    # type: (Union[str, Path, Image.Image, bytes, io.BytesIO], **Any) -> str
    """Converts image to WebP data-uri."""
    opts = SdkOptions(**options) if options else sdk_opts

    quality = opts.image_preview_quality

    if isinstance(img, bytes):
        img = io.BytesIO(img)

    if not isinstance(img, Image.Image):
        img = Image.open(img)

    raw = io.BytesIO()
    img.save(raw, format="WEBP", lossless=False, quality=quality, method=6)
    raw.seek(0)
    enc = base64.b64encode(raw.read()).decode("ascii")
    return "data:image/webp;base64," + enc


def normalize_image(imo, **options):
    # type: (Image.Image, **bool) -> Sequence[int]
    """Normalize image for hash calculation.

    :param Image.Image imo: Pillow Image Object
    :key bool image_transpose: Transpose image according to EXIF Orientation
    :key bool image_fill: Add gray background to image if it has alpha transparency
    :key bool image_trim: Crop empty borders of image
    :return: Normalized and flattened image 1024-pixel array (from 32x32 gray pixels)
    :rtype: Sequence[int]
    """
    opts = SdkOptions(**options) if options else sdk_opts

    # Transpose image according to EXIF Orientation tag
    if opts.image_transpose:
        imo = transpose_image(imo)

    # Add gray background to image if it has alpha transparency
    if opts.image_fill:
        imo = fill_image(imo)

    # Trim uniform colored (empty) border if there is one
    if opts.image_trim:
        imo = trim_image(imo)

    # Convert to grayscale
    imo = imo.convert("L")

    # Resize to 32x32
    im = imo.resize((32, 32), Image.BICUBIC)

    # A flattened sequence of grayscale pixel values (1024 pixels)
    pixels = im.getdata()

    return pixels


def transpose_image(imo):
    # type: (Image.Image) -> Image.Image
    """
    Transpose image according to EXIF Orientation tag

    :param Image.Image imo: Pillow Image Object
    :return: EXIF transposed image
    :rtype: Image.Image
    """
    imo = ImageOps.exif_transpose(imo)
    logger.debug(f"Image exif transpose applied")
    return imo


def fill_image(imo):
    # type: (Image.Image) -> Image.Image
    """
    Add gray background to image if it has alpha transparency

    :param Image.Image imo: Pillow Image Object
    :return: Image where alpha transparency is replaced with gray background.
    :rtype: Image.Image
    """
    if imo.mode in ("RGBA", "LA") or (imo.mode == "P" and "transparency" in imo.info):
        if imo.mode != "RGBA":
            imo = imo.convert("RGBA")
        bg = Image.new("RGBA", imo.size, (126, 126, 126))
        imo = Image.alpha_composite(bg, imo)
        logger.debug(f"Image transparency filled with gray")
    return imo


def trim_image(imo):
    # type: (Image.Image) -> Image.Image
    """Trim uniform colored (empty) border.

    Takes the upper left pixel as reference to

    :param Image.Image imo: Pillow Image Object
    :return: Image with uniform colored (empty) border removed.
    :rtype: Image.Image
    """

    bg = Image.new(imo.mode, imo.size, imo.getpixel((0, 0)))
    diff = ImageChops.difference(imo, bg)
    diff = ImageChops.add(diff, diff)
    bbox = diff.getbbox()
    if bbox:
        logger.debug(f"Image has been trimmed")
        return imo.crop(bbox)
    return imo



# -*- coding: utf-8 -*-
from json import JSONDecodeError
import json
import subprocess
from more_itertools import chunked
from iscc.bin import fpcalc_bin, fpcalc_install
from iscc.schema import FeatureType, Readable
from iscc.options import SdkOptions
from iscc_core.codec import encode_base64
from typing import Any, List, Union
from iscc import uread
import tinytag
import iscc_core
from loguru import logger as log
from iscc.schema import File, Uri


def extract_audio_metadata(file):
    # type: (Union[File, Uri]) -> dict

    infile = uread.open_data(file)
    if not hasattr(infile, "name"):
        log.warning("Cannot extract audio metadata without file.name")
        return dict()
    file_path = infile.name

    try:
        tag = tinytag.TinyTag.get(file_path)
    except Exception as e:
        log.error(f"failed audio metadata extraction: {e}")
        return dict()
    result = iscc_core.ContentCodeAudio(iscc="dummy")
    result.title = tag.title
    if tag.duration:
        result.duration = round(float(tag.duration), 3)
    else:
        log.warning("failed to extract duration")

    return dict(sorted(result.dict(exclude={"iscc"}).items()))


def extract_audio_features(data, **options):
    # type: (Readable, **Any) -> dict
    """Returns Chromaprint fingerprint.

    A dictionary with keys:
    - duration: total duration of extracted fingerprint in seconds (Broken with pipe)
    - fingerprint: 32-bit (4 byte) integers as features
    """
    opts = SdkOptions(**options)
    length = str(opts.audio_max_duration)
    ufile = uread.open_data(data)
    cmd = [fpcalc_bin(), "-raw", "-json", "-signed", "-length", length]
    if hasattr(ufile, "name"):
        cmd.append(ufile.name)
        data = None
    else:
        data = ufile.read()
        cmd.append("-")

    try:
        res = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)
    except FileNotFoundError:
        fpcalc_install()
        res = subprocess.run(cmd, input=data, stdout=subprocess.PIPE, check=True)

    output = res.stdout.decode("utf-8")
    try:
        result = json.loads(output)
    except JSONDecodeError:
        raise RuntimeError(f'Fpcalc decode error: {res.stderr.decode("utf-8")}')

    return result


def encode_audio_features(features):
    # type: (List[int]) -> dict
    """Pack into 64-bit base encoded features.

    Note: Last feature will be 32-bit if number of features is uneven.
    """
    if len(features) % 2 != 0:
        features.append(0)
    fingerprints = []
    for int_features in chunked(features, 2):
        digest = b""
        for int_feature in int_features:
            digest += int_feature.to_bytes(4, "big", signed=True)
        fingerprints.append(encode_base64(digest))
    return dict(kind=FeatureType.audio.value, version=0, features=fingerprints)



# -*- coding: utf-8 -*-
"""Fuctione for Data-Code granular fingerprints."""
from loguru import logger as log
from typing import Any, List
from codetiming import Timer
from more_itertools import chunked
from iscc_core.minhash import minhash_64
from iscc_core.codec import encode_base64
from iscc.schema import FeatureType
from iscc.options import SdkOptions


def encode_data_features(sizes, features, **options):
    # type: (List[int], List[int], **Any) -> dict
    """Reduce and encode low level features.

    :return: dicttionary with {"sizes": ..., "features": ...}
    """
    opts = SdkOptions(**options)
    with Timer(text="{:0.4f}s for data features encoding", logger=log.debug):
        encoded_sizes = [sum(fh) for fh in chunked(sizes, opts.data_granular_factor)]
        encoded_features = [
            encode_base64(minhash_64(cf))
            for cf in chunked(features, opts.data_granular_factor)
        ]
    return dict(
        kind=FeatureType.data.value,
        version=0,
        features=encoded_features,
        sizes=encoded_sizes,
    )



# -*- coding: utf-8 -*-
"""Wrap iscc-core functions for more convenient inputs/outputs"""
from typing import Iterable, List
from iscc_core import codec


def decompose(iscc_code):
    # type: (codec.AnyISCC) -> List[codec.Code]
    """
    Wrap iscc_core.codec.decompose to handle more input types and return higher
    level Code object instead of plain ISCC strings.
    """
    if isinstance(iscc_code, str):
        iscc_code = codec.clean(iscc_code)
    code_obj = codec.Code(iscc_code)
    components = [codec.Code(c) for c in codec.decompose(code_obj.code)]
    return components



# -*- coding: utf-8 -*-
from loguru import logger
from typing import List, Optional, Union
import mimetypes
import magic
from PIL import Image
from iscc.schema import GMT, Readable
from iscc import uread
from loguru import logger as log

__all__ = [
    "mime_guess",
    "mime_normalize",
    "mime_supported",
    "mime_clean",
    "mime_to_gmt",
    "mime_from_name",
    "mime_from_data",
]


def mime_guess(data, file_name=None):
    # type: (Readable, str) -> str
    """Heuristic guessing of mediatype for different kinds of inputs.

    We try matching by file extension. If that fails we match by content sniffing.
    """

    guess_name, guess_data = None, None
    file = uread.open_data(data)

    if file_name is None:
        if hasattr(file, "name"):
            file_name = file.name
        elif hasattr(file, "filename"):
            file_name = file.filename

    if file_name:
        guess_name = mime_from_name(file_name)

    guess_data = mime_from_data(file.read(4096))

    # Normalize
    guess_data = mime_normalize(guess_data)
    guess_name = mime_normalize(guess_name)
    media_type = guess_name or guess_data

    # Special cases of missdetection
    if guess_data and "ogg" in guess_data:
        media_type = guess_data

    log.debug(f"{media_type} media-type detected")

    return media_type


def mime_normalize(mime: str) -> str:
    """Return normalized version of a mediatype."""
    return MEDIATYPE_NORM.get(mime, mime)


# TODO make use of mime_supported
def mime_supported(mime: str) -> bool:
    """Check if mediatype is supported"""
    return mime_normalize(mime) in SUPPORTED_MEDIATYPES


def mime_from_name(name: str) -> Optional[str]:
    """Guess mediatype from filename or url."""
    return mimetypes.guess_type(name)[0]


def mime_from_data(data: bytes) -> Optional[str]:
    """Guess mediatype by sniffing raw header data."""
    return magic.from_buffer(data, mime=True)


def mime_clean(mime: Union[str, List]):
    """
    Clean mimetype/content-type string or first entry of a list of mimetype strings.
    Also removes semicolon separated encoding information.
    """
    if mime and isinstance(mime, List):
        mime = mime[0]
    if mime:
        mime = mime.split(";")[0]
    return mime.strip()


def mime_to_gmt(mime_type: str, file_path=None):
    """Get generic mediatype from mimetype."""
    mime_type = mime_clean(mime_type)
    if mime_type == "image/gif" and file_path:
        img = Image.open(file_path)
        if img.is_animated:
            return "video"
        else:
            return "image"
    entry = SUPPORTED_MEDIATYPES.get(mime_type)
    if entry:
        return entry["gmt"]
    gmt = mime_type.split("/")[0]
    if gmt in list(GMT):
        logger.warning(f"Guessing GMT from {mime_type}")
        return gmt


mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/x-mobipocket-ebook", ".mobi")
mimetypes.add_type("application/x-sqlite3", ".sqlite")
mimetypes.add_type("video/mp4", ".f4v")
mimetypes.add_type("video/h264", ".h264")


SUPPORTED_MEDIATYPES = {
    # Text Formats
    "application/rtf": {"gmt": "text", "ext": "rtf"},
    "application/msword": {"gmt": "text", "ext": "doc"},
    "application/pdf": {"gmt": "text", "ext": "pdf"},
    "application/epub+zip": {"gmt": "text", "ext": "epub"},
    "text/xml": {"gmt": "text", "ext": "xml"},
    "application/json": {"gmt": "text", "ext": "json"},
    "application/xhtml+xml": {"gmt": "text", "ext": "xhtml"},
    "application/vnd.oasis.opendocument.text": {"gmt": "text", "ext": "odt"},
    "text/html": {"gmt": "text", "ext": "html"},
    "text/plain": {"gmt": "text", "ext": "txt"},
    "application/x-ibooks+zip": {"gmt": "text", "ext": "ibooks"},
    "text/markdown": {"gmt": "text", "ext": ["md", "markdown"]},
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": {
        "gmt": "text",
        "ext": "docx",
    },
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": {
        "gmt": "text",
        "ext": "xlsx",
    },
    # Note: pptx only detected by file extension. Sniffing gives 'application/zip'
    "application/vnd.openxmlformats-officedocument.presentationml.presentation": {
        "gmt": "text",
        "ext": "pptx",
    },
    "application/vnd.ms-excel": {"gmt": "text", "ext": "xls"},
    "application/x-mobipocket-ebook": {
        "gmt": "text",
        "ext": ["mobi", "prc", "azw", "azw3", "azw4"],
    },
    # Image Formats
    "image/bmp": {"gmt": "image", "ext": "bmp"},
    "image/gif": {"gmt": "image", "ext": "gif"},
    "image/jpeg": {"gmt": "image", "ext": ["jpg", "jpeg"]},
    "image/png": {"gmt": "image", "ext": "png"},
    "image/tiff": {"gmt": "image", "ext": "tif"},
    "image/vnd.adobe.photoshop": {"gmt": "image", "ext": "psd"},
    "application/postscript": {"gmt": "image", "ext": "eps"},
    # Audio Formats
    "audio/mpeg": {"gmt": "audio", "ext": "mp3"},
    "audio/wav": {"gmt": "audio", "ext": "wav"},
    "audio/x-wav": {"gmt": "audio", "ext": "wav"},
    "audio/ogg": {"gmt": "audio", "ext": "ogg"},
    "audio/aiff": {"gmt": "audio", "ext": "aif"},
    "audio/x-aiff": {"gmt": "audio", "ext": "aif"},
    "audio/x-flac": {"gmt": "audio", "ext": "flac"},
    "audio/opus": {"gmt": "audio", "ext": "opus"},
    # Video Formats
    "application/vnd.rn-realmedia": {"gmt": "video", "ext": "rm"},
    "video/x-dirac": {"gmt": "video", "ext": "drc"},
    "video/3gpp": {"gmt": "video", "ext": "3gp"},
    "video/3gpp2": {"gmt": "video", "ext": "3g2"},
    "video/x-ms-asf": {"gmt": "video", "ext": "asf"},
    "video/avi": {"gmt": "video", "ext": "avi"},
    "video/webm": {"gmt": "video", "ext": "webm"},
    "video/mpeg": {"gmt": "video", "ext": ["mpeg", "mpg", "m1v", "vob"]},
    "video/mp4": {"gmt": "video", "ext": "mp4"},
    "video/x-m4v": {"gmt": "video", "ext": "m4v"},
    "video/x-matroska": {"gmt": "video", "ext": "mkv"},
    "video/ogg": {"gmt": "video", "ext": ["ogg", "ogv"]},
    "video/quicktime": {"gmt": "video", "ext": ["mov", "f4v"]},
    "video/x-flv": {"gmt": "video", "ext": "flv"},
    "application/x-shockwave-flash": {"gmt": "video", "ext": "swf"},
    "video/h264": {"gmt": "video", "ext": "h264"},
    "video/x-ms-wmv": {"gmt": "video", "ext": "wmv"},
}

MEDIATYPE_NORM = {
    "audio/x-aiff": "audio/aiff",
    "audio/x-wav": "audio/wav",
    "image/x-ms-bmp": "image/bmp",
    "video/x-msvideo": "video/avi",
}

SUPPORTED_EXTENSIONS = []
for v in SUPPORTED_MEDIATYPES.values():
    ext = v["ext"]
    if isinstance(ext, str):
        SUPPORTED_EXTENSIONS.append(ext)
    else:
        for e in ext:
            SUPPORTED_EXTENSIONS.append(e)



# -*- coding: utf-8 -*-
from os.path import join
from pathlib import Path
from typing import Optional
from iscc_core.options import CoreOptions
from pydantic import Field
from iscc import APP_DIR


class SdkOptions(CoreOptions):
    """Options for ISCC generation"""

    class Config:
        env_file = "iscc-sdk.env"
        env_file_encoding = "utf-8"

    all_granular: Optional[bool] = Field(
        None,
        description="Generate granular fingerprints for all code types "
        "(overides individual options).",
    )

    all_preview: Optional[bool] = Field(
        None,
        description="Generate previews for all code types "
        "(overrides individual options).",
    )

    meta_title_from_uri: bool = Field(
        True,
        description="Use normalized filename as title if we have nor explicit title "
        "and also no title from metadata extraction.",
    )

    text_granular: bool = Field(
        True, description="Calculate and return granular text features"
    )

    text_avg_chunk_size: int = Field(
        1024,
        description="Avg number of characters per text chunk for granular fingerprints",
    )

    text_guess_title: bool = Field(
        True,
        description="Use first line from text as title if we don´t have an explicit title",
    )

    text_store: bool = Field(
        False,
        description="Store extracted paintext (with filename: <datahash>.txt.gz).",
    )

    image_transpose: bool = Field(
        True,
        description="Transpose image according to EXIF Orientation tag",
    )

    image_fill: bool = Field(
        True, description="Add gray background to image if it has alpha transparency"
    )

    image_trim: bool = Field(True, description="Crop empty borders of images")

    image_granular: bool = Field(
        True, description="Calculate and return granular image features"
    )

    image_granular_n: int = Field(32, description="Top-N granular features to retain")

    image_preview: bool = Field(True, description="Generate image preview thumbnail")

    image_preview_size: int = Field(
        96, description="Size of larger side of thumbnail in pixels"
    )

    image_preview_quality: int = Field(
        30, description="Image compression setting (0-100)"
    )

    audio_granular: bool = Field(
        True, description="Calculate and return granular audio features"
    )

    audio_max_duration: int = Field(
        0,
        description="Maximum seconds of audio to process (0 for full fingerprint)",
    )

    video_fps: int = Field(
        5,
        description="Frames per second to process for video hash (ignored when 0).",
    )

    video_crop: bool = Field(
        True, description="Detect and remove black borders before processing"
    )

    video_granular: bool = Field(True, description="Generate granular features")

    video_scenes: bool = Field(
        True, description="Use scene detection for granular features"
    )

    video_scenes_ffmpeg: bool = Field(
        True,
        description="Use ffmpeg built-in scene detection. (Less accurate than "
        "pyscenedetect but at least 2x faster).",
    )

    video_scenes_ffmpeg_th: float = Field(0.4, description="FFMPEG scene cut threshold")

    video_scenes_fs: int = Field(
        2,
        description="Number of frames to skip per processing step "
        "for scene detection with pyscenedetect. Higher values will increase detection "
        "speed and decrease detection quality (PySceneDetect only).",
    )

    video_scenes_th: int = Field(
        40,
        description="Threshold for scene detection. Higher values detect less scenes "
        "(PySceneDetect only).",
    )

    video_scenes_min: int = Field(
        15,
        description="Minimum number of frames per scene. (PySceneDetect only)",
    )

    video_window: int = Field(
        7,
        description="Seconds of video per granular feature if using rolling window mode",
    )

    video_overlap: int = Field(
        3, description="Seconds of video that overlap if using rolling window mode"
    )

    video_include_fingerprint: bool = Field(
        False, description="Include raw MPEG-7 Video Signature in output"
    )

    video_preview: bool = Field(True, description="Generate video preview thumbnail(s)")

    video_hwaccel: Optional[str] = Field(
        None,
        description="Use ffmpeg hardware acceleration for video processing "
        "(use the string `auto` as value to activate).",
    )

    data_granular: bool = Field(False, description="Return granular data features.")

    data_granular_factor: int = Field(
        64, description="Size of granular data chunks time average chunk size"
    )

    index_root: Path = Field(join(APP_DIR, "db"), description="Storage root path")

    index_components: bool = Field(
        True, description="Create inverted index of component to ISCCs"
    )

    index_features: bool = Field(
        False, description="Create inverted index of features to ISCCs"
    )

    index_metadata: bool = Field(False, description="Store metadata in index")


sdk_opts = SdkOptions()



# -*- coding: utf-8 -*-
import operator
from enum import Enum
import mmap
from io import BufferedReader, BytesIO
from pathlib import Path
from typing import BinaryIO, List, Optional, Union
from pydantic import BaseModel, Field, validator, conint, confloat, constr
from iscc.metrics import distance_b64


DEFAULT_WINDOW = 7
DEFAULT_OVERLAP = 3
FEATURE_REGEX = "^[-A-Za-z0-9_]{11}"
IKEY_REGEX = "^[-A-Za-z0-9+/]*={0,2}$"


Data = Union[bytes, bytearray, memoryview]
Uri = Union[str, Path]
File = Union[BinaryIO, mmap.mmap, BytesIO, BufferedReader]
Readable = Union[Uri, Data, File]
PositiveNum = Union[conint(strict=True, ge=0), confloat(strict=True, ge=0.0)]
IndexKey = Union[
    conint(strict=True, ge=0),
    constr(strict=True, min_length=1, max_length=64, regex=IKEY_REGEX),
]


class GMT(str, Enum):
    """Generic Metdia Type"""

    text = "text"
    image = "image"
    audio = "audio"
    video = "video"
    unknown = "unknown"


class FeatureMatch(BaseModel):
    """A single granular feature match result."""

    class Config:
        frozen = True

    key: Optional[IndexKey] = Field(description="Unique key of ISCC entry.")
    matched_iscc: str = Field(description="The matched (candidate) ISCC")
    kind: str = Field(
        description="The kind of feature that has been matched.",
    )
    source_feature: str = Field(description="The original feature that was queried.")
    source_pos: Optional[PositiveNum] = Field(
        description="The position of of the original feature"
    )
    matched_feature: str = Field(description="The feature hash of the matched entry.")
    matched_position: Optional[PositiveNum] = Field(
        description="The position of the feature in the matched content."
    )
    distance: Optional[int] = Field(description="The hamming distance of the match")

    @validator("distance", always=True)
    def calculate_distance(cls, v, values):
        if v is None:
            v = distance_b64(values["source_feature"], values["matched_feature"])
        return v

    def __lt__(self, other):
        """Adds support for deterministic sorting by distance."""
        a = (
            self.distance,
            self.matched_iscc,
            self.matched_position,
            self.matched_feature,
        )
        b = (
            other.distance,
            other.matched_iscc,
            other.matched_position,
            other.matched_feature,
        )
        return operator.lt(a, b)


class IsccMatch(BaseModel):
    """Metrics of a matched ISCC."""

    class Config:
        frozen = True

    key: Optional[IndexKey] = Field(description="Unique key of ISCC entry.")
    matched_iscc: str = Field(description="The ISCC found to match with the query.")
    distance: Optional[int] = Field(description="Hamming distance of the full code")
    mdist: Optional[int] = Field(description="Hamming distance of Meta-Code.")
    cdist: Optional[int] = Field(description="Hamming distance of Content-Code.")
    ddist: Optional[int] = Field(description="Hamming distance of Data-Code.")
    imatch: Optional[bool] = Field(description="Wether Instance-Code is identical.")

    def __lt__(self, other):
        """Adds support for deterministic sorting by distance."""
        a = self.distance, self.matched_iscc
        b = other.distance, other.matched_iscc
        return operator.lt(a, b)


class QueryResult(BaseModel):

    iscc_matches: Optional[List[IsccMatch]] = Field(
        default_factory=list, description="Matched ISCCs with distance metrics"
    )
    feature_matches: Optional[List[FeatureMatch]] = Field(
        default_factory=list, description="ISCCs matched by granular features."
    )


class FeatureType(str, Enum):
    """Type of granular features."""

    text = "text"
    image = "image"
    audio = "audio"
    video = "video"
    data = "data"


class Features(BaseModel):
    """Granular feature codes.

    If only a list of features is provided it is assumed that those have been created
    with the default values for 'window' and 'overlap'.

    If sizes are provided it is assumed that we deal with custom segment sizes
    based on content aware chunking.
    """

    kind: FeatureType = Field(description="Type of granular features")
    version: int = Field(description="Version of feature generation algorithm")
    features: List[str] = Field(
        description="Segmentwise 64-bit features (base64url encoded).",
        regex=FEATURE_REGEX,
        min_items=1,
    )
    sizes: Optional[List[PositiveNum]] = Field(
        description="Sizes of segmets used for feature calculation.",
        min_items=1,
    )
    window: Optional[int] = Field(
        DEFAULT_WINDOW,
        description="Window size of feature segments",
    )
    overlap: Optional[int] = Field(
        DEFAULT_OVERLAP,
        description="Overlap size of feature segments",
    )

    # TODO add positions or locations array

    @property
    def type_id(self) -> str:
        """Composite string of type-version used as table name for feature indexing."""
        return f"{self.kind}-{self.version}"


class ISCC(BaseModel):
    class Config:
        extra = "forbid"
        validate_assignment = True

    version: str = Field(
        "0-0-0",
        title="ISCC Schema Version",
        description="Version of ISCC Metadata Schema (SchemaVer).",
        const=True,
    )
    iscc: str = Field(description="ISCC code of the identified digital asset.")

    # Essential Metadata
    title: Optional[str] = Field(
        description="The title or name of the intangible creation manifested by the"
        " identified digital asset"
    )
    extra: Optional[str] = Field(
        description="Descriptive, industry-sector or use-case specific metadata (used "
        "as immutable input for Meta-Code generation). Any text string "
        "(including json or json-ld) indicative of the identity of the "
        "referent may be used."
    )

    # binary: bool = Field(description="Extra metadata was supplied in binary format.")

    # File Properties
    filename: Optional[str] = Field(
        description="Filename of the referenced digital asset (automatically used as "
        "fallback if no seed_title element is specified)"
    )
    filesize: Optional[int] = Field(description="File size of media asset in bytes.")
    mediatype: Optional[str] = Field(description="IANA Media Type (MIME type)")

    # Cryptographic hashes
    tophash: Optional[str] = Field(
        title="tophash",
        description="Blake3 hash over concatenation of metahash and datahash",
    )
    metahash: Optional[str] = Field(
        title="metahash", description="Blake3 hash of metadata."
    )
    datahash: Optional[str] = Field(
        title="datahash", description="Blake3 hash of media file."
    )

    gmt: GMT = Field(GMT.unknown, description="Generic Media Type")

    # Audio Visual Media
    duration: Optional[float] = Field(
        description="Duration of audio-visual media in secondes."
    )
    fps: Optional[float] = Field(description="Frames per second of video assets.")
    width: Optional[int] = Field(description="Width of visual media in pixels.")
    height: Optional[int] = Field(description="Height of visual media in pixels.")

    # Textual Media
    characters: Optional[int] = Field(
        description="Number of text characters (code points after Unicode "
        "normalization) (GMT Text only)."
    )

    language: Optional[Union[str, List[str]]] = Field(
        description="Language(s) of content (BCP-47) in weighted order."
    )

    # Granular Features
    features: Optional[List[Features]] = Field(
        description="Standardized fingerprints for granular content "
        "recognition and matching purposes."
    )

    # Presentational Metadata
    preview: Optional[str] = Field(description="Uri of media asset preview.")

    # Todo: fingerprint should be a repeatable complex object (maybe part of features)
    fingerprint: Optional[str] = Field(
        description="Base64 encoded original raw fingerprint (MPEG-7, Chromaprint) "
        "from which features and/or codes have been derived."
    )


if __name__ == "__main__":
    """Save ISCC JSON schema"""
    from os.path import abspath, dirname, join

    HERE = dirname(abspath(__file__))
    SCHEMA_PATH = join(HERE, "iscc.json")
    schema = ISCC.schema_json()
    with open(SCHEMA_PATH, "wt", encoding="UTF-8") as outf:
        outf.write(ISCC.schema_json(indent=2))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iscc']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.4,<9.0',
 'annoy>=1.17,<2.0',
 'bidict>=0.21,<0.22',
 'codetiming>=1.3,<2.0',
 'humanize>=3.12,<4.0',
 'install-jdk>=0.3.0,<0.4.0',
 'iscc-core>=0.1,<0.2',
 'jmespath>=0.10,<0.11',
 'langcodes>=3.3,<4.0',
 'langdetect>=1.0,<2.0',
 'lmdb>=1.2,<2.0',
 'mobi>=0.3,<0.4',
 'msgpack>=1.0,<2.0',
 'numpy==1.19.3',
 'pyexiv2==2.7.1',
 'requests>=2.26,<3.0',
 'tinytag>=1.6,<2.0']

extras_require = \
{':sys_platform == "linux"': ['python-magic>=0.4.22,<0.5.0'],
 ':sys_platform == "win32" or sys_platform == "darwin"': ['python-magic-bin>=0.4.14,<0.5.0'],
 'optional': ['scenedetect>=0.5.5,<0.6.0',
              'opencv-contrib-python-headless>=4.5,<5.0']}

setup_kwargs = {
    'name': 'iscc',
    'version': '1.1.0b14',
    'description': 'ISCC: Reference Implementation',
    'long_description': '# ISCC - Spec and Reference Code\n\n[![Build](https://github.com/iscc/iscc-specs/actions/workflows/test.yml/badge.svg?branch=version-1.1)](https://github.com/iscc/iscc-specs/actions/workflows/test.yml)\n[![License](https://img.shields.io/pypi/l/iscc.svg)](https://pypi.python.org/pypi/iscc/)\n[![Downloads](https://pepy.tech/badge/iscc)](https://pepy.tech/project/iscc)\n[![DOI](https://zenodo.org/badge/96668860.svg)](https://zenodo.org/badge/latestdoi/96668860)\n\nThe **International Standard Content Code** is a proposal for an [open standard](https://en.wikipedia.org/wiki/Open_standard) for decentralized content identification. This repository contains the specification of the proposed **ISCC Standard** and a reference implementation in Python3. The latest published version of the specification can be found at [iscc.codes](https://iscc.codes)\n\n| NOTE: This is ISCC Version 1.1 work in progress!!! |\n| --- |\n\n## Installing the reference code\n\nThe reference code is published with the package name [iscc](https://pypi.python.org/pypi/iscc) on Python Package Index. Install it with:\n\n``` bash\npip install iscc==1.1.0b14\n```\n\nIf your system is setup to compile c-extensions install with:\n\n``` bash\npip install iscc[turbo]==1.1.0b14\n```\n\nThis will install cython and build binary extansions for faster ISCC processing.\n\nTo install the required binaries for content extraction do:\n\n```python\nfrom iscc.bin import install\ninstall()\n```\n\n## Using the reference code\n\nA short example on how to create an ISCC Code with the reference implementation.\n\n``` python\n>>> import iscc\n>>> iscc.code_iscc("README.md", all_granular=True)\n{\'version\': \'0-0-0\',\n \'iscc\': \'KADYHLZUJ43U3LX7G7PNLS54JHAET3ANW4EQ3YQIP3LDAZHEYIS5GWI\',\n \'title\': \'# ISCC Spec and Reference Code\',\n \'filename\': \'README.md\',\n \'filesize\': 3840,\n \'mediatype\': \'text/markdown\',\n \'tophash\': \'00194e2c4e5570e637bd18667740fdcf7f1683d6ccace7f5c0cc6531f6e982e5\',\n \'metahash\': \'828dd01bf76b78fc448f6d2ab25008835d2993c6acde205235dc942083c4677d\',\n \'datahash\': \'d63064e4c225d3594bdf60c30bcb04554e53059d9077a6f330f8295b4420ded5\',\n \'gmt\': <GMT.text: \'text\'>,\n \'characters\': 3457,\n \'language\': \'en\',\n \'features\': [{\'kind\': <FeatureType.data: \'data\'>,\n               \'version\': 0,\n               \'features\': [\'7A23CQ3iCH4\'],\n               \'sizes\': [3840]},\n              {\'kind\': <FeatureType.text: \'text\'>,\n               \'version\': 0,\n               \'features\': [\'Nt6V67hJxmk\',\n                            \'9HvPYqt1rQw\',\n                            \'ld1FLbp7A50\',\n                            \'M8aTn6atuB0\'],\n               \'sizes\': [2340, 309, 292, 516]}]}\n```\n\n## Working with the specification\n\n| NOTE: This is ISCC Version 1.1 - The specs are not yet updated!!! |\n| --- |\n\nThe entire **ISCC Specification** is written in plain text [Markdown](https://en.wikipedia.org/wiki/Markdown). The markdown content is than built and published with the excellent [mkdocs](http://www.mkdocs.org/) documetation tool. If you have some basic command line skills you can build and run the specification site on your own computer. Make sure you have the [git](https://git-scm.com/) and [Python](https://www.python.org/) and [Poetry](https://pypi.org/project/poetry/) installed on your system and follow these steps on the command line:\n\n``` bash\ngit clone https://github.com/iscc/iscc-specs.git\ncd iscc-specs\npoetry install\nmkdocs serve\n```\n\nAll specification documents can be found in the `./docs` subfolder or the repository. The recommended editor for the markdown files is [Typora](https://typora.io/). If you have commit rights to the [main repository](https://github.com/iscc/iscc-specs) you can deploy the site with a simple `mkdocs gh-deploy`.\n\n## Contribute\n\nPull requests and other contributions are welcome. Use the [Github Issues](https://github.com/iscc/iscc-specs/issues) section of this project to discuss ideas for the **ISCC Specification**. You may also want  join our developer chat on Telegram at <https://t.me/iscc_dev>.\n\n## License\n\nAll of documentation is licensed under the [CC BY-NC-SA 4.0](https://creativecommons.org/licenses/by-nc-sa/4.0/).\n\nReference code is licensed under BSD-2-Clause.\n',
    'author': 'Titusz Pan',
    'author_email': 'tp@py7.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://iscc.codes/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<3.10',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Print application directory. Usefull for CI.

Usage:

$python -m iscc.app_dir
"""
import iscc


def main():
    print(iscc.APP_DIR)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import json
from fractions import Fraction
from codetiming import Timer
from iscc_core.code_content_video import soft_hash_video_v0
from loguru import logger as log
import subprocess
from pathlib import Path
from tempfile import mkdtemp
import os
import sys
from subprocess import Popen, PIPE, DEVNULL
from secrets import token_hex
from typing import Any, Generator, List, Tuple, Optional, Union
from statistics import mode
from langcodes import standardize_tag
from iscc.options import SdkOptions
from iscc import uread
from iscc.schema import FeatureType, Readable, File, Uri
from iscc_core.codec import encode_base64
from iscc.mp7 import Frame
from iscc.bin import ffmpeg_bin, ffprobe_bin
import jmespath
import iscc_core


FFMPEG = ffmpeg_bin()
FFPROBE = ffprobe_bin()
Scene = Union[Tuple["FrameTimecode", "FrameTimecode"], Tuple[float, float]]
SceneSig = Tuple[List[str], List[int]]  # feature hashes, scene durations


def extract_video_metadata(file):
    # type: (Union[File, Uri]) -> dict

    infile = uread.open_data(file)
    if not hasattr(infile, "name"):
        log.warning("Cannot extract video metadata without file.name")
        return dict()
    file_path = infile.name

    cmd = [
        FFPROBE,
        "-hide_banner",
        "-loglevel",
        "fatal",
        "-find_stream_info",
        "-show_error",
        "-show_format",
        "-show_streams",
        "-show_programs",
        "-show_chapters",
        "-show_private_data",
        "-print_format",
        "json",
        "-i",
        file_path,
    ]
    res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    metadata = json.loads(res.stdout)

    video_streams = jmespath.search("streams[?codec_type=='video']", metadata)
    n_streams = len(video_streams)
    if n_streams == 0:
        raise ValueError("No video stream detected")
    if n_streams != 1:
        log.warning(f"Detected {n_streams} video streams.")

    vstream = video_streams[0]

    result = iscc_core.ContentCodeVideo(iscc="dummy")

    # Duration
    duration_format = jmespath.search("format.duration", metadata)
    duration_streams = jmespath.search(
        "streams[?codec_type=='video'].duration", metadata
    )
    durations = (
        [duration_format] + duration_streams if duration_format else duration_streams
    )
    if durations:
        duration = max(round(float(d), 3) for d in durations)
        result.duration = duration
    else:
        log.warning("failed to extract duration")

    # Dimensions
    result.height = vstream.get("height")
    result.width = vstream.get("width")

    # FPS
    fps = vstream.get("r_frame_rate")
    if fps:
        fps = round(float(Fraction(fps)), 3)
        result.fps = fps

    # Title
    result.title = jmespath.search("format.tags.title", metadata)

    # Language
    langs = jmespath.search("streams[*].tags.language", metadata)
    lang = jmespath.search("format.tags.language", metadata)
    if lang:
        langs.insert(0, lang)
    if langs:
        result.language = standardize_tag(langs[0])

    return result.dict(exclude={"iscc"})


def extract_video_preview(file, **options):
    # type: (Union[File, Uri], **Any) -> Optional[bytes]
    """Extract thumbnail from video and return raw png byte data."""
    opts = SdkOptions(**options)
    size = opts.image_preview_size
    infile = uread.open_data(file)

    if not hasattr(infile, "name"):
        log.warning("Cannot extract preview without file.name")
        return None

    file_path = infile.name

    cmd = [
        FFMPEG,
        "-i",
        file_path,
        "-vf",
        f"thumbnail,scale={size}:-1",
        "-frames:v",
        "1",
        "-c:v",
        "png",
        "-f",
        "image2pipe",
        "-",
    ]
    with Timer(text="{:0.4f}s for video preview extraction", logger=log.debug):
        result = subprocess.run(cmd, stdout=PIPE, stderr=DEVNULL)
    return result.stdout


def extract_video_signature(uri, crop=None, **options):
    # type: (Union[File, Uri], Optional[str], **Any) -> bytes
    """Extract MPEG-7 Video Signature

    :param uri: File to process
    :param crop: FFMPEG style cropsting "w:h:x:y"
    :key video_fps: Frames per second for signature processing
    :key video_hwaccel: Hadware acceleration mode (None or "auto")
    :return: raw signature data
    """

    opts = SdkOptions(**options)

    infile = uread.open_data(uri)
    if not hasattr(infile, "name"):
        log.error("Cannot extract signature without file.name")
        raise ValueError(f"Cannot extract signature from {type(infile)}")

    infile_path = infile.name
    sigfile_path = Path(mkdtemp(), token_hex(16) + ".bin")
    sigfile_path_escaped = sigfile_path.as_posix().replace(":", "\\\\:")

    vf = f"signature=format=binary:filename={sigfile_path_escaped}"
    if crop:
        vf = f"{crop}," + vf

    if opts.video_fps:
        vf = f"fps=fps={opts.video_fps}," + vf

    cmd = [FFMPEG]

    if opts.video_hwaccel is not None:
        cmd.extend(["-hwaccel", opts.video_hwaccel])

    cmd.extend(["-i", infile_path, "-vf", vf, "-f", "null", "-"])

    log.debug(f"video sig extraction with {cmd}")
    with Timer(text="{:0.4f}s for video sig extraction", logger=log.debug):
        subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    with open(sigfile_path, "rb") as sig:
        sigdata = sig.read()
    os.remove(sigfile_path)
    return sigdata


def extract_video_signature_cutpoints(uri, crop=None, **options):
    # type: (Union[File, Uri], Optional[str], **Any) -> Tuple[bytes, List[Scene]]
    """Extract MPEG-7 Video Signature together with ffmpeg scdet cutponts

    :param uri: File to process
    :param crop: FFMPEG style cropsting "w:h:x:y"
    :key video_fps: Frames per second for signature processing
    :key video_hwaccel: Hadware acceleration mode (None or "auto")
    :return: raw signature data
    """

    opts = SdkOptions(**options)

    infile = uread.open_data(uri)
    if not hasattr(infile, "name"):
        log.error("Cannot extract signature without file.name")
        raise ValueError(f"Cannot extract signature from {type(infile)}")

    infile_path = infile.name
    sig_path = Path(mkdtemp(), token_hex(16) + ".bin")
    sig_path_escaped = sig_path.as_posix().replace(":", "\\\\:")
    scene_path = Path(mkdtemp(), token_hex(16) + ".cut")
    scene_path_escaped = scene_path.as_posix().replace(":", "\\\\:")

    sig = f"signature=format=binary:filename={sig_path_escaped}"
    if crop:
        sig = f"{crop}," + sig

    if opts.video_fps:
        sig = f"fps=fps={opts.video_fps}," + sig

    scene = f"select='gte(scene,0)',metadata=print:file={scene_path_escaped}"

    cmd = [
        FFMPEG,
    ]

    if opts.video_hwaccel is not None:
        cmd.extend(["-hwaccel", opts.video_hwaccel])

    cmd.extend(
        [
            "-i",
            infile_path,
            "-an",
            "-sn",
            "-filter_complex",
            f"split[in1][in2];[in1]{scene}[out1];[in2]{sig}[out2]",
            "-map",
            "[out1]",
            "-f",
            "null",
            "-",
            "-map",
            "[out2]",
            "-f",
            "null",
            "-",
        ]
    )

    log.debug(f"video sig and cutpoint extraction with {subprocess.list2cmdline(cmd)}")
    with Timer(text="{:0.4f}s for video sig and cutpoint extraction", logger=log.debug):
        subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    with open(sig_path, "rb") as sigin:
        sigdata = sigin.read()
    with open(scene_path, "rt", encoding="utf-8") as scenein:
        scenetext = scenein.read()
    os.remove(sig_path)
    os.remove(scene_path)
    scenes = parse_ffmpeg_scenes(scenetext, **options)
    return sigdata, scenes


def compute_video_features_rolling(frames, **options):
    # type: (List[Frame], **int) -> dict
    """
    Compute video signatures based on rolling window.

    Generates segment-wise features where 'window' is the duration of segments in
    seconds and 'overlap' is the number of seconds that overlap for each segment.
    """
    opts = SdkOptions(**options)
    window = opts.video_window
    overlap = opts.video_overlap
    assert overlap < window, "Overlap must be shorter than window"
    shift = window - overlap
    cut_indexes = [0]
    start = 0
    for fidx, frame in enumerate(frames):
        if frame.elapsed > start + shift:
            cut_indexes.append(fidx)
            start = frame.elapsed
    sigs = []
    for ci in cut_indexes:
        segment_frames = []
        start = frames[ci].elapsed
        for frame in frames[ci:]:
            segment_frames.append(tuple(frame.vector.tolist()))
            if frame.elapsed > start + window:
                sigs.append(encode_base64(soft_hash_video_v0(segment_frames, bits=64)))
                break
    return dict(
        kind=FeatureType.video.value, features=sigs, window=window, overlap=overlap
    )


def detect_video_crop(uri):
    # type: (Union[Uri, File]) -> str
    """
    Detect crop value for video.
    Example result: crop=176:96:0:24
    """

    infile = uread.open_data(uri)
    if not hasattr(infile, "name"):
        raise ValueError(f"Cannot extract signature for {type(infile)}")
    file_path = infile.name

    cmd = [
        FFMPEG,
        "-t",
        "3:00",
        "-i",
        file_path,
        "-vf",
        "cropdetect",
        "-f",
        "null",
        "-",
    ]
    res = subprocess.run(cmd, stderr=subprocess.PIPE)
    text = res.stderr.decode(encoding=sys.stdout.encoding)
    crops = [
        line.split()[-1]
        for line in text.splitlines()
        if line.startswith("[Parsed_cropdetect")
    ]
    return mode(crops)


def detect_video_scenes(uri, **options):
    # type: (Union[Uri, File], **Any) -> List[float]
    """Compute Scenedetection and return cutpoints.

    :param uri: Video file to be processed (file path or object)
    :key video_scenes_th: Threshold for scene detection. Higher value -> less scenes.
    :key video_scenes_fs: Frame Skip, number of frames to skip per processing step.
    :key video_scenes_min: Minimum number of frames per scene.
    :key video_scenes_previews: Generate per scene preview thumbnails.
    :return: List of tuples with start end end FrameTimecode.
    """
    try:
        from scenedetect import (
            ContentDetector,
            FrameTimecode,
            SceneManager,
            VideoManager,
        )
    except ImportError:
        raise EnvironmentError(
            "Please install `scenedetect`python module for advanced scenedetection."
        )

    opts = SdkOptions(**options)

    infile = uread.open_data(uri)
    if not hasattr(infile, "name"):
        raise ValueError(f"Cannot extract signature for {type(infile)}")
    file_path = infile.name

    video_manager = VideoManager([file_path])
    scene_manager = SceneManager()
    scene_manager.add_detector(
        ContentDetector(
            threshold=opts.video_scenes_th,
            min_scene_len=opts.video_scenes_min,
        )
    )
    base_timecode = video_manager.get_base_timecode()
    video_manager.set_downscale_factor()
    video_manager.start()

    with Timer(text="{:0.4f}s for video scene detection", logger=log.debug):
        scene_manager.detect_scenes(
            frame_source=video_manager,
            show_progress=False,
            frame_skip=opts.video_scenes_fs,
        )
    # Use end frames from scene_list to exclude 0 and include last frame time
    scenes = scene_manager.get_scene_list(base_timecode)
    cutlist = [round(float(scene[1].get_seconds()), 3) for scene in scenes]
    return cutlist


def _signature_extractor(crop=None):
    # type: (Optional[str]) -> Generator
    """Streaming signature generator (use gen.send(chunk)).
    WARNING: Not up to date
    """

    def raw_generator(crop=None):
        # type: (Optional[str]) -> Generator
        sigfile = Path(mkdtemp(), token_hex(16) + ".bin")
        # We need to escape colon from windows drive letter prefixes
        # See: https://stackoverflow.com/a/28770642/51627
        sigfile_escaped = sigfile.as_posix().replace(":", "\\\\:")
        if crop:
            vf = "{},signature=format=binary:filename={}".format(crop, sigfile_escaped)
        else:
            vf = "signature=format=binary:filename={}".format(sigfile_escaped)
        cmd = [FFMPEG, "-i", "-", "-vf", vf, "-f", "null", "-"]
        proc = Popen(cmd, stdout=DEVNULL, stderr=DEVNULL, stdin=PIPE)
        data = yield
        while data:
            proc.stdin.write(data)
            data = yield
        proc.stdin.close()
        proc.wait()
        with open(sigfile, "rb") as infile:
            sigdata = infile.read()
        os.remove(sigfile)
        yield sigdata

    initialized_generator = raw_generator(crop)
    next(initialized_generator)
    return initialized_generator


def parse_ffmpeg_scenes(scene_text, **options):
    # type: (str, **Any) -> List[float]
    """Parse scene score output from ffmpeg

    :param str scene_text: Scene score output from ffmpeg
    :key float video_scenes_ffmpeg_th: Scene Score threshold for valid cutpoint
    """
    if not scene_text.strip():
        return []

    opts = SdkOptions(**options)
    times = []
    scores = []
    for line in scene_text.splitlines():
        if line.startswith("frame:"):
            ts = round(float(line.split()[-1].split(":")[-1]), 3)
            times.append(ts)
        if line.startswith("lavfi.scene_score"):
            scores.append(float(line.split("=")[-1]))

    cutpoints = []
    for ts, score in zip(times, scores):
        if score >= opts.video_scenes_ffmpeg_th:
            cutpoints.append(ts)

    # append last frame timestamp if not in cutpoints
    if cutpoints and cutpoints[-1] != times[-1]:
        cutpoints.append(times[-1])

    return cutpoints[1:]


def compute_video_features_scenes(frames, scenes):
    # type: (List[Frame], List[float]) -> dict
    """Compute video signatures for individual scenes in video.
    Returns a dictionary conforming to `shema.Feature`- objects.
    """
    features, sizes, segment = [], [], []
    start_frame = 0
    for cidx, cutpoint in enumerate(scenes):
        try:
            frames = frames[start_frame:]
        except IndexError:
            break
        for fidx, frame in enumerate(frames):
            frame_t = tuple(frame.vector.tolist())
            segment.append(frame_t)
            if frame.elapsed >= cutpoint:
                features.append(encode_base64(soft_hash_video_v0(segment, 64)))
                segment = []
                prev_cutpoint = 0 if cidx == 0 else scenes[cidx - 1]
                duration = round(cutpoint - prev_cutpoint, 3)
                sizes.append(duration)
                start_frame = fidx + 1
                break
    if not features:
        log.info("No scenes detected. Use all frames")
        segment = [tuple(frame.vector.tolist()) for frame in frames]
        features = [encode_base64(soft_hash_video_v0(segment, 64))]
        sizes = [round(float(frames[-1].elapsed), 3)]

    return dict(kind=FeatureType.video.value, version=0, features=features, sizes=sizes)
